#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from .expressions import TypedExpr, SequenceExpr, Scalar, \
    BooleanSequenceExpr, BooleanScalar, CollectionExpr, Expr, \
    int_number_sequences, int_number_scalars
from .core import NodeMetaclass
from . import utils
from . import errors
from .. import types
from ..utils import FunctionWrapper
from ...compat import six


class AnyOpNodeMetaClass(NodeMetaclass):
    def __new__(mcs, name, bases, kv):
        if '_add_args_slots' not in kv:
            kv['_add_args_slots'] = False

        return super(AnyOpNodeMetaClass, mcs).__new__(mcs, name, bases, kv)


class AnyOp(six.with_metaclass(AnyOpNodeMetaClass, TypedExpr)):
    __slots__ = ()

    @classmethod
    def _get_type(cls, *args, **kwargs):
        if '_data_type' in kwargs:
            return SequenceExpr._get_type(cls, *args, **kwargs)
        else:
            return Scalar._get_type(cls, *args, **kwargs)

    @classmethod
    def _typed_classes(cls, *args, **kwargs):
        if '_data_type' in kwargs:
            return SequenceExpr._typed_classes(cls, *args, **kwargs)
        else:
            return Scalar._typed_classes(cls, *args, **kwargs)

    @classmethod
    def _base_class(cls, *args, **kwargs):
        if '_data_type' in kwargs:
            return SequenceExpr
        else:
            return Scalar

    @classmethod
    def is_seq(cls):
        return issubclass(cls, SequenceExpr)


class ElementWise(AnyOp):
    __slots__ = ()
    _args = '_input',

    @property
    def node_name(self):
        return self.__class__.__name__

    @property
    def name(self):
        return self._name or self._input.name

    @property
    def input(self):
        return self._input

    def iter_args(self):
        for it in zip(['_input'] + [arg.lstrip('_') for arg in self._args[1:]],
                      self.args):
            yield it


class ElementOp(ElementWise):
    __slots__ = ()

    def accept(self, visitor):
        return visitor.visit_element_op(self)


class MappedExpr(ElementWise):
    _slots = '_func', '_func_args', '_func_kwargs', '_resources', \
             '_multiple', '_raw_inputs'
    _args = '_inputs', '_collection_resources'
    node_name = 'Map'

    def _init(self, *args, **kwargs):
        self._init_attr('_multiple', False)
        self._init_attr('_raw_inputs', None)
        super(MappedExpr, self)._init(*args, **kwargs)

    @property
    def inputs(self):
        return self._inputs

    @property
    def name(self):
        if self._name is not None:
            return self._name
        if len(self._inputs) == 1:
            return self._inputs[0].name

    @property
    def source_name(self):
        if self._source_name is not None:
            return self._source_name
        if len(self._inputs) == 1:
            return self._inputs[0].source_name

    @property
    def input_types(self):
        return [it.dtype for it in self._inputs]

    @property
    def raw_input_types(self):
        if self._raw_inputs:
            return [it.dtype for it in self._raw_inputs]
        return self.input_types

    @property
    def func(self):
        return self._func

    @func.setter
    def func(self, f):
        self._func = f

    def accept(self, visitor):
        return visitor.visit_function(self)


class IsNull(ElementOp):
    __slots__ = ()


class IsNa(ElementOp):
    __slots__ = ()


class NotNull(ElementOp):
    __slots__ = ()


class NotNa(ElementOp):
    __slots__ = ()


class FillNa(ElementOp):
    _args = '_input', '_fill_value'

    def _init(self, *args, **kwargs):
        self._init_attr('_fill_value', None)

        super(FillNa, self)._init(*args, **kwargs)

        if self._fill_value is not None and not isinstance(self._fill_value, Expr):
            tp = types.validate_value_type(self._fill_value)
            if not self.input.dtype.can_implicit_cast(tp):
                raise ValueError('fillna cannot cast value from %s to %s' % (
                    tp, self.input.dtype))
            self._fill_value = Scalar(_value=self._fill_value, _value_type=self.dtype)
        if not self.input.dtype.can_implicit_cast(self._fill_value.dtype):
            raise ValueError('fillna cannot cast value from %s to %s' % (
                self._fill_value.dtype, self.input.dtype))

    @property
    def fill_value(self):
        if self._fill_value is None:
            return
        if isinstance(self._fill_value, Scalar):
            return self._fill_value.value
        return self._fill_value


class IsIn(ElementOp):
    _args = '_input', '_values',

    def _init(self, *args, **kwargs):
        super(IsIn, self)._init(*args, **kwargs)

        self._values = _scalar(self._values, tp=self._input.dtype)
        if isinstance(self._values, list):
            self._values = tuple(self._values)
        if not isinstance(self._values, tuple):
            if not isinstance(self._values, SequenceExpr):
                raise ValueError('isin accept iterable object or sequence')
            self._values = (self._values, )

    @property
    def name(self):
        return self._name


class NotIn(ElementOp):
    _args = '_input', '_values',

    def _init(self, *args, **kwargs):
        super(NotIn, self)._init(*args, **kwargs)

        self._values = _scalar(self._values, tp=self._input.dtype)
        if isinstance(self._values, list):
            self._values = tuple(self._values)
        if not isinstance(self._values, tuple):
            if not isinstance(self._values, SequenceExpr):
                raise ValueError('notin accept iterable object or sequence')
            self._values = (self._values, )

    @property
    def name(self):
        return self._name


class Between(ElementOp):
    _args = '_input', '_left', '_right', '_inclusive'

    def _init(self, *args, **kwargs):
        self._init_attr('_left', None)
        self._init_attr('_right', None)
        self._init_attr('_inclusive', None)

        super(Between, self)._init(*args, **kwargs)

        for attr in self._args[1:]:
            val = getattr(self, attr)
            if val is not None and not isinstance(val, Expr):
                setattr(self, attr, Scalar(_value=val))

    def _get_val(self, attr):
        val = getattr(self, attr)
        if val is None:
            return
        if isinstance(val, Scalar):
            return val.value
        return val

    @property
    def left(self):
        return self._get_val('_left')

    @property
    def right(self):
        return self._get_val('_right')

    @property
    def inclusive(self):
        return self._get_val('_inclusive')


class IfElse(ElementOp):
    _args = '_input', '_then', '_else'

    @property
    def name(self):
        return self._name


class Switch(ElementOp):
    _args = '_input', '_case', '_conditions', '_thens', '_default'

    def iter_args(self):
        def _names():
            if hasattr(self, '_case'):
                yield 'case'
            for _ in self._conditions:
                yield 'when'
                yield 'then'
            yield 'default'

        def _args():
            if hasattr(self, '_case'):
                yield self._case
            for condition, then in zip(self._conditions, self._thens):
                yield condition
                yield then
            yield self._default

        for it in zip(_names(), _args()):
            yield it

    @property
    def name(self):
        return self._name


class Cut(ElementOp):
    _args = '_input', '_bins', '_right', '_labels', '_include_lowest', \
            '_include_under', '_include_over'

    def _init(self, *args, **kwargs):
        super(Cut, self)._init(*args, **kwargs)

        if len(self._bins) == 0:
            raise ValueError('Must be at least one bin edge')
        elif len(self._bins) == 1:
            if not self._include_under or not self._include_over:
                raise ValueError('If one bin edge provided, must have'
                                 ' include_under=True and include_over=True')

        for arg in self._args[1:]:
            obj = getattr(self, arg)
            setattr(self, arg, _scalar(obj))

        under = 1 if self._include_under.value else 0
        over = 1 if self._include_over.value else 0
        size = len(self._bins) - 1 + under + over

        if self._labels is not None and len(self._labels) != size:
            raise ValueError('Labels size must be exactly the size of bins')
        if self._labels is None:
            self._labels = _scalar(list(range(size)))

    @property
    def right(self):
        return self._right.value

    @property
    def include_lowest(self):
        return self._include_lowest.value

    @property
    def include_under(self):
        return self._include_under.value

    @property
    def include_over(self):
        return self._include_over.value

    @property
    def name(self):
        return self._name


class IntToDatetime(ElementOp):
    __slots__ = ()


class Func(AnyOp):
    _slots = '_func_name', '_rtype'
    _args = '_inputs',

    def accept(self, visitor):
        visitor.visit_function(self)


class FuncFactory(object):
    def __getattr__(self, item):
        if not isinstance(item, six.string_types):
            raise TypeError('Function name should be provided, expect str, got %s' % type(item))

        def gen_func(*args, **kwargs):
            func_name = item
            project = kwargs.pop('project', None)
            if project:
                func_name = '%s:%s' % (project, item)

            expr_name = kwargs.pop('name', None)
            rtype = kwargs.pop('rtype', types.string)
            rtype = types.validate_data_type(rtype)
            is_seq = kwargs.pop('seq', None)

            args = tuple(arg if isinstance(arg, Expr) else Scalar(_value=arg) for arg in args)
            is_seq = is_seq if is_seq is not None else any(isinstance(arg, SequenceExpr) for arg in args)
            kw = {'_value_type': rtype} if not is_seq else {'_data_type': rtype}

            if expr_name is None:
                exprs = tuple(arg for arg in args if isinstance(arg, SequenceExpr))
                if len(exprs) == 1:
                    expr_name = exprs[0].name

            if expr_name:
                return Func(_func_name=func_name, _inputs=args, **kw).rename(expr_name)
            else:
                return Func(_func_name=func_name, _inputs=args, **kw)

        return gen_func


def _map(expr, func, rtype=None, resources=None, args=(), **kwargs):
    """
    Call func on each element of this sequence.

    :param func: lambda, function, :class:`odps.models.Function`,
                 or str which is the name of :class:`odps.models.Funtion`
    :param rtype: if not provided, will be the dtype of this sequence
    :return: a new sequence

    :Example:

    >>> df.id.map(lambda x: x + 1)
    """

    name = None
    if isinstance(func, FunctionWrapper):
        if func.output_names:
            if len(func.output_names) > 1:
                raise ValueError('Map column has more than one name')
            name = func.output_names[0]
        if func.output_types:
            rtype = rtype or func.output_types[0]
        func = func._func

    if rtype is None:
        rtype = utils.get_annotation_rtype(func)

    from ...models import Function

    rtype = rtype or expr.dtype
    output_type = types.validate_data_type(rtype)

    if isinstance(func, six.string_types):
        pass
    elif isinstance(func, Function):
        pass
    elif inspect.isclass(func):
        pass
    elif not callable(func):
        raise ValueError('`func` must be a function or a callable class')

    collection_resources = utils.get_collection_resources(resources)

    is_seq = isinstance(expr, SequenceExpr)
    if is_seq:
        return MappedExpr(_data_type=output_type, _func=func, _inputs=[expr, ],
                          _func_args=args, _func_kwargs=kwargs, _name=name,
                          _resources=resources, _collection_resources=collection_resources)
    else:
        return MappedExpr(_value_type=output_type, _func=func, _inputs=[expr, ],
                          _func_args=args, _func_kwargs=kwargs, _name=name,
                          _resources=resources, _collection_resources=collection_resources)


def _hash(expr, func=None):
    """
    Calculate the hash value.

    :param expr:
    :param func: hash function
    :return:
    """
    if func is None:
        func = lambda x: hash(x)

    return _map(expr, func=func, rtype=types.int64)


def _isnull(expr):
    """
    Return a sequence or scalar according to the input indicating if the values are null.

    :param expr: sequence or scalar
    :return: sequence or scalar
    """

    if isinstance(expr, SequenceExpr):
        return IsNull(_input=expr, _data_type=types.boolean)
    elif isinstance(expr, Scalar):
        return IsNull(_input=expr, _value_type=types.boolean)


def _isna(expr):
    """
    Return a sequence or scalar according to the input indicating if the values are null.

    :param expr: sequence or scalar
    :return: sequence or scalar
    """

    if isinstance(expr, SequenceExpr):
        return IsNa(_input=expr, _data_type=types.boolean)
    elif isinstance(expr, Scalar):
        return IsNa(_input=expr, _value_type=types.boolean)


def _notnull(expr):
    """
    Return a sequence or scalar according to the input indicating if the values are not null.

    :param expr: sequence or scalar
    :return: sequence or scalar
    """

    if isinstance(expr, SequenceExpr):
        return NotNull(_input=expr, _data_type=types.boolean)
    elif isinstance(expr, Scalar):
        return NotNull(_input=expr, _value_type=types.boolean)


def _notna(expr):
    """
    Return a sequence or scalar according to the input indicating if the values are not null.

    :param expr: sequence or scalar
    :return: sequence or scalar
    """

    if isinstance(expr, SequenceExpr):
        return NotNa(_input=expr, _data_type=types.boolean)
    elif isinstance(expr, Scalar):
        return NotNa(_input=expr, _value_type=types.boolean)


def _fillna(expr, value):
    """
    Fill null with value.

    :param expr: sequence or scalar
    :param value: value to fill into
    :return: sequence or scalar
    """

    if isinstance(expr, SequenceExpr):
        return FillNa(_input=expr, _fill_value=value, _data_type=expr.dtype)
    elif isinstance(expr, Scalar):
        return FillNa(_input=expr, _fill_value=value, _value_type=expr.dtype)


def _isin(expr, values):
    """
    Return a boolean sequence or scalar showing whether
    each element is exactly contained in the passed `values`.

    :param expr: sequence or scalar
    :param values: `list` object or sequence
    :return: boolean sequence or scalar
    """

    from .merge import _make_different_sources

    if isinstance(values, SequenceExpr):
        expr, values = _make_different_sources(expr, values)

    if isinstance(expr, SequenceExpr):
        return IsIn(_input=expr, _values=values, _data_type=types.boolean)
    elif isinstance(expr, Scalar):
        return IsIn(_input=expr, _values=values, _value_type=types.boolean)


def _notin(expr, values):
    """
    Return a boolean sequence or scalar showing whether
    each element is not contained in the passed `values`.

    :param expr: sequence or scalar
    :param values: `list` object or sequence
    :return: boolean sequence or scalar
    """

    if isinstance(expr, SequenceExpr):
        return NotIn(_input=expr, _values=values, _data_type=types.boolean)
    elif isinstance(expr, Scalar):
        return NotIn(_input=expr, _values=values, _value_type=types.boolean)


def _between(expr, left, right, inclusive=True):
    """
    Return a boolean sequence or scalar show whether
    each element is between `left` and `right`.

    :param expr: sequence or scalar
    :param left: left value
    :param right: right value
    :param inclusive: if true, will be left <= expr <= right, else will be left < expr < right
    :return: boolean sequence or scalar
    """

    if isinstance(expr, SequenceExpr):
        return Between(_input=expr, _left=left, _right=right,
                       _inclusive=inclusive, _data_type=types.boolean)
    elif isinstance(expr, Scalar):
        return Between(_input=expr, _left=left, _right=right,
                       _inclusive=inclusive, _value_type=types.boolean)


def _scalar(val, tp=None):
    if val is None:
        return
    if isinstance(val, Expr):
        return val
    if isinstance(val, (tuple, list)):
        return type(val)(_scalar(it, tp=tp) for it in val)
    else:
        return Scalar(_value=val, _value_type=tp)


def _ifelse(expr, true_expr, false_expr):
    """
    Given a boolean sequence or scalar, if true will return the left, else return the right one.

    :param expr: sequence or scalar
    :param true_expr:
    :param false_expr:
    :return: sequence or scalar

    :Example:

    >>> (df.id == 3).ifelse(df.id, df.fid.astype('int'))
    >>> df.isMale.ifelse(df.male_count, df.female_count)
    """

    tps = (SequenceExpr, Scalar)
    if not isinstance(true_expr, tps):
        true_expr = Scalar(_value=true_expr)
    if not isinstance(false_expr, tps):
        false_expr = Scalar(_value=false_expr)

    output_type = utils.highest_precedence_data_type(
            *[true_expr.dtype, false_expr.dtype])
    is_sequence = isinstance(expr, SequenceExpr) or \
                  isinstance(true_expr, SequenceExpr) or \
                  isinstance(false_expr, SequenceExpr)

    if is_sequence:
        return IfElse(_input=expr, _then=true_expr, _else=false_expr,
                      _data_type=output_type)
    else:
        return IfElse(_input=expr, _then=true_expr, _else=false_expr,
                      _value_type=output_type)


def _switch(expr, *args, **kw):
    """
    Similar to the case-when in SQL. Refer to the example below

    :param expr:
    :param args:
    :param kw:
    :return: sequence or scalar

    :Example:

    >>> # if df.id == 3 then df.name
    >>> # elif df.id == df.fid.abs() then df.name + 'test'
    >>> # default: 'test'
    >>> df.id.switch(3, df.name, df.fid.abs(), df.name + 'test', default='test')
    """
    default = _scalar(kw.get('default'))

    if len(args) <= 0:
        raise errors.ExpressionError('Switch must accept more than one condition')

    if all(isinstance(arg, tuple) and len(arg) == 2 for arg in args):
        conditions, thens = [list(tp) for tp in zip(*args)]
    else:
        conditions = [arg for i, arg in enumerate(args) if i % 2 == 0]
        thens = [arg for i, arg in enumerate(args) if i % 2 == 1]

    if len(conditions) == len(thens):
        conditions, thens = _scalar(conditions), _scalar(thens)
    else:
        raise errors.ExpressionError('Switch should be called by case and then pairs')

    if isinstance(expr, (Scalar, SequenceExpr)):
        case = expr
    else:
        case = None
        if not all(hasattr(it, 'dtype') and it.dtype == types.boolean for it in conditions):
            raise errors.ExpressionError('Switch must be called by all boolean conditions')

    res = thens if default is None else thens + [default, ]
    output_type = utils.highest_precedence_data_type(*(it.dtype for it in res))

    is_seq = isinstance(expr, SequenceExpr) or \
        any(isinstance(it, SequenceExpr) for it in conditions) or \
        any(isinstance(it, SequenceExpr) for it in res)
    if case is not None:
        is_seq = is_seq or isinstance(case, SequenceExpr)

    kwargs = dict()
    if is_seq:
        kwargs['_data_type'] = output_type
    else:
        kwargs['_value_type'] = output_type
    return Switch(_input=expr, _case=case, _conditions=conditions,
                  _thens=thens, _default=default, **kwargs)


def switch(*args, **kwargs):
    return _switch(None, *args, **kwargs)


def _cut(expr, bins, right=True, labels=None, include_lowest=False,
         include_under=False, include_over=False):
    """
    Return indices of half-open bins to which each value of `expr` belongs.

    :param expr: sequence or scalar
    :param bins: list of scalars
    :param right: indicates whether the bins include the rightmost edge or not. If right == True(the default),
                  then the bins [1, 2, 3, 4] indicate (1, 2], (2, 3], (3, 4]
    :param labels: Usesd as labes for the resulting bins. Must be of the same length as the resulting bins.
    :param include_lowest: Whether the first interval should be left-inclusive or not.
    :param include_under: include the bin below the leftmost edge or not
    :param include_over: include the bin above the rightmost edge or not
    :return: sequence or scalar
    """

    is_seq = isinstance(expr, SequenceExpr)
    dtype = utils.highest_precedence_data_type(
        *(types.validate_value_type(it) for it in labels)) \
        if labels is not None else types.int64
    kw = {}
    if is_seq:
        kw['_data_type'] = dtype
    else:
        kw['_value_type'] = dtype

    return Cut(_input=expr, _bins=bins, _right=right, _labels=labels,
               _include_lowest=include_lowest, _include_under=include_under,
               _include_over=include_over, **kw)


def _int_to_datetime(expr):
    """
    Return a sequence or scalar that is the datetime value of the current numeric sequence or scalar.

    :param expr: sequence or scalar
    :return: sequence or scalar
    """

    if isinstance(expr, SequenceExpr):
        return IntToDatetime(_input=expr, _data_type=types.datetime)
    elif isinstance(expr, Scalar):
        return IntToDatetime(_input=expr, _value_type=types.datetime)


_element_methods = dict(
    map=_map,
    isnull=_isnull,
    notnull=_notnull,
    isna=_isna,
    notna=_notna,
    fillna=_fillna,
    between=_between,
    switch=_switch,
    cut=_cut,
    isin=_isin,
    notin=_notin,
    hash=_hash,
)

utils.add_method(SequenceExpr, _element_methods)
utils.add_method(Scalar, _element_methods)

BooleanSequenceExpr.ifelse = _ifelse
BooleanScalar.ifelse = _ifelse

CollectionExpr.switch = _switch

for int_number_sequence in int_number_sequences + int_number_scalars:
    int_number_sequence.to_datetime = _int_to_datetime



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import re
from collections import namedtuple

from ..compat import six, getargspec
from .. import ODPS, options, utils
from ..inter import list_rooms

PROJECT_REGEX = re.compile(r'.*project *= *(?P<project>[^\(\),]+)')
NAME_REGEX = re.compile(r'.*name *= *(?P<name>[^\(\),]+)')
TEMP_TABLE_PREFIXES = [
    utils.TEMP_TABLE_PREFIX,
    'jdbc_temp_tbl_',
    'pai_temp_',
    'temp_xlib_table_'
]


class RoomCompleter(object):
    def __init__(self, ipython=None):
        self._ipython = ipython
        self._regex_str = r'^%(enter|setup|teardown|stores) +'

    def register(self):
        self._ipython.set_hook('complete_command', self, re_key=self._regex_str)

    def __call__(self, completer, event):
        cursor_text = event.text_until_cursor
        _, prefix = cursor_text.split(' ', 1)
        prefix = prefix.strip()
        rooms = [n for n in list_rooms() if n.startswith(prefix)]
        return rooms[:options.completion_size]


class BaseCompleter(object):
    def __init__(self, ipython=None):
        self._ipython = ipython
        self._regex_str = self.build_regex()

    def build_regex(self):
        pass

    def get_list_call(self, cursor_str, full_line=None):
        pass

    def register(self):
        self._ipython.set_hook('complete_command', self, re_key=self._regex_str)

    def __call__(self, completer, event):
        cursor_text = event.text_until_cursor
        full_line = event.line
        call_tuple = self.get_list_call(cursor_text, full_line)
        if call_tuple is None:
            return None
        code, quote = call_tuple
        if quote is None:
            quote = '\''
        else:
            quote = ''

        try:
            list_gen = self._ipython.ev(code)
        except:
            return None

        def _is_temp_table(tn):
            return any(tn.startswith(p) for p in TEMP_TABLE_PREFIXES)

        def render_object(o):
            if hasattr(o, 'name'):
                if _is_temp_table(o.name):
                    return None
                name = o.name
            else:
                name = str(o)
            return quote + utils.str_to_printable(name, auto_quote=False) + quote

        names = [render_object(o) for idx, o in enumerate(list_gen)
                 if idx < options.completion_size]
        return [n for n in names if n]


class ObjectCompleter(BaseCompleter):
    @staticmethod
    def iter_methods():
        odps_methods = [m_name for m_name in dir(ODPS) if callable(getattr(ODPS, m_name))
                        and not m_name.startswith('_')]
        odps_listers = [m_name for m_name in odps_methods if m_name.startswith('list_')]

        for m_name in odps_methods:
            for prefix in ('get_', 'delete_', 'write_', 'read_'):
                if m_name.startswith(prefix):
                    lister = None
                    lister_prefix = m_name.replace(prefix, 'list_')
                    for l in odps_listers:
                        if l.startswith(lister_prefix):
                            lister = l
                            break
                    if lister:
                        yield m_name, lister
                    break

    def build_regex(self):
        self._methods = {}
        method_type = namedtuple('MethodType', 'use_prefix list_method')

        for m_name, lister in self.iter_methods():
            arg_tuple = getargspec(getattr(ODPS, lister))
            use_prefix = 'prefix' in arg_tuple[0]
            self._methods[m_name] = method_type(use_prefix=use_prefix, list_method=lister)

        _regex_str = (
            r'(^|.*[\(\)\s,=]+)(?P<odps>[^\(\)\s,]+)\.(?P<getfn>'
            + '|'.join(six.iterkeys(self._methods))
            + r')\('
        )
        self._regex = re.compile(_regex_str + r'(?P<args>[^\(\)]*)$')
        return _regex_str

    def get_list_call(self, cursor_str, full_line=None):
        full_line = full_line or cursor_str

        cmatch = self._regex.match(cursor_str)
        if cmatch is None:
            return None
        odps_obj = cmatch.group('odps')
        get_cmd = cmatch.group('getfn')
        arg_str = cmatch.group('args').strip()

        project = 'None'
        arg_start, arg_cursor = cmatch.span('args')
        arg_body = full_line[arg_start:]
        pmatch = PROJECT_REGEX.match(arg_body)
        if pmatch:
            project = pmatch.group('project')

        nmatch = NAME_REGEX.match(arg_str)
        name_str = nmatch.group('name') if nmatch else arg_str

        quote = None
        if name_str != '' and not (name_str.startswith('\'') or name_str.startswith('\"')):
            return None
        if name_str.endswith('\"') or name_str.endswith('\''):
            return None
        if name_str:
            quote = name_str[0]
            name_str = name_str[1:]
        if '"' in name_str or "'" in name_str:
            return None

        if name_str and self._methods[get_cmd].use_prefix:
            formatter = '{odps}.{func}(prefix="{prefix}", project={project})'
        else:
            formatter = '{odps}.{func}(project={project})'
        return formatter.format(odps=odps_obj, func=self._methods[get_cmd].list_method,
                                prefix=name_str, project=project), quote


def load_ipython_extension(ipython):
    ObjectCompleter(ipython).register()
    RoomCompleter(ipython).register()



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

from ..expr.exporters import get_ml_input, get_input_field_names, get_input_field_ids, get_input_field_id
from ..expr.mixin import ml_collection_mixin
from ...compat import Enum


class TextFieldRole(Enum):
    DOC_ID = 'DOC_ID'
    DOC_CONTENT = 'DOC_CONTENT'
    WORD = 'WORD'
    SENTENCE = 'SENTENCE'
    WORD_COUNT = 'WORD_COUNT'


@ml_collection_mixin
class TextDFMixIn(object):
    __slots__ = ()

    field_role_enum = TextFieldRole
    non_feature_roles = set([TextFieldRole.DOC_ID, ])

"""
Common NLP exporters
"""
get_doc_id_column = partial(get_input_field_names, field_role=TextFieldRole.DOC_ID)
get_doc_content_column = partial(get_input_field_names, field_role=TextFieldRole.DOC_CONTENT)
get_sentence_column = partial(get_input_field_names, field_role=TextFieldRole.SENTENCE)
get_word_column = partial(get_input_field_names, field_role=TextFieldRole.WORD)
get_word_count_column = partial(get_input_field_names, field_role=TextFieldRole.WORD_COUNT)

get_doc_content_column_ids = partial(get_input_field_ids, field_role=TextFieldRole.DOC_CONTENT)
get_doc_content_column_id = partial(get_input_field_id, field_role=TextFieldRole.DOC_CONTENT)


def normalize_get_append_col_names(expr, param_name, input_name):
    if expr._params[param_name]:
        return expr._params[param_name]
    data_obj = get_ml_input(expr, input_name)
    if data_obj is None:
        return None
    fields = data_obj._ml_fields
    sel_cols = set(expr._params['selectedColNames'].split(',')) if expr._params['selectedColNames']\
        else set(get_doc_content_column(expr, param_name, input_name))
    return [f.name for f in fields if f.name not in sel_cols]



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

include "util_c.pxi"

from libc.stdint cimport *
from libc.string cimport *

from .wire_format import TAG_TYPE_BITS as PY_TAG_TYPE_BITS, _TAG_TYPE_MASK as _PY_TAG_TYPE_MASK

cdef:
    int TAG_TYPE_BITS = PY_TAG_TYPE_BITS
    int _TAG_TYPE_MASK = _PY_TAG_TYPE_MASK
    size_t _BUFFER_SIZE = 64 * 1024
    size_t _MIN_SERIALIZED_INT_SIZE = 10  # ceil(64 / 7)


cdef class CDecoder:

    def __cinit__(self, stream):
        self._pos = 0
        self._stream = stream
        self._buffer = stream.read(_BUFFER_SIZE)
        self._begin = self._buffer
        self._end = self._begin + len(self._buffer)
        self._is_source_eof = False

    cdef int32_t read_field_number(self) except? -1 nogil:
        if self._end - self._begin < _MIN_SERIALIZED_INT_SIZE:
            self._load_next_buffer()

        cdef int32_t tag_and_type
        tag_and_type = self.read_uint32()
        return tag_and_type >> TAG_TYPE_BITS

    cdef read_field_number_and_wire_type(self):
        if self._end - self._begin < _MIN_SERIALIZED_INT_SIZE:
            self._load_next_buffer()

        cdef int32_t tag_and_type
        tag_and_type = self.read_uint32()
        return (tag_and_type >> TAG_TYPE_BITS), (tag_and_type & _TAG_TYPE_MASK)

    cdef size_t position(self) nogil:
        return self._pos

    cdef int32_t read_sint32(self) except? -1 nogil:
        if self._end - self._begin < _MIN_SERIALIZED_INT_SIZE:
            self._load_next_buffer()
        return get_signed_varint32(&self._begin, self._end, &self._pos)

    cdef uint32_t read_uint32(self) except? 0xffffffff nogil:
        if self._end - self._begin < _MIN_SERIALIZED_INT_SIZE:
            self._load_next_buffer()
        return get_varint32(&self._begin, self._end, &self._pos)

    cdef int64_t read_sint64(self) except? -1 nogil:
        if self._end - self._begin < _MIN_SERIALIZED_INT_SIZE:
            self._load_next_buffer()
        return get_signed_varint64(&self._begin, self._end, &self._pos)

    cdef uint64_t read_uint64(self) except? 0xffffffff nogil:
        if self._end - self._begin < _MIN_SERIALIZED_INT_SIZE:
            self._load_next_buffer()
        return get_varint64(&self._begin, self._end, &self._pos)

    cdef bint read_bool(self) except? False nogil:
        if self._end - self._begin < _MIN_SERIALIZED_INT_SIZE:
            self._load_next_buffer()
        return get_varint32(&self._begin, self._end, &self._pos)

    cdef double read_double(self) except? -1.0 nogil:
        cdef double retval

        if self._end - self._begin < sizeof(double):
            self._load_next_buffer()

        memcpy(&retval, self._begin, sizeof(double))
        self._begin += sizeof(double)
        self._pos += sizeof(double)
        return retval

    cdef float read_float(self) except? -1.0 nogil:
        cdef float retval

        if self._end - self._begin < sizeof(float):
            self._load_next_buffer()
        memcpy(&retval, self._begin, sizeof(float))
        self._begin += sizeof(float)
        self._pos += sizeof(float)
        return retval

    cdef bytes read_string(self):
        cdef size_t need

        if self._end - self._begin < _MIN_SERIALIZED_INT_SIZE:
            self._load_next_buffer()
        need = self.read_uint32()

        cdef int offset
        cdef list result = []
        cdef bytes chunk
        while need > 0:
            if need > self._end - self._begin:
                self._load_next_buffer()
            if need > self._end - self._begin:
                # _load_next_buffer() will put unused buffer at head after invoked
                chunk = self._buffer
                self._pos += self._end - self._begin
                need -= self._end - self._begin
                self._begin = self._end
                result.append(chunk)
                continue

            offset = len(self._buffer) - (self._end - self._begin)
            chunk = self._buffer[offset:offset + need]
            self._begin += need
            self._pos += need
            result.append(chunk)
            need = 0

        if len(result) == 0:
            return b''
        elif len(result) == 1:
            return result[0]
        else:
            return b''.join(result)

    cdef int _load_next_buffer(self) except -1 with gil:
        if self._is_source_eof and (self._begin >= self._end):
            raise EOFError

        cdef bytes data = self._stream.read(_BUFFER_SIZE)
        cdef size_t length = len(data)
        if length == 0:
            self._is_source_eof = True
            return 0

        cdef bytes left
        if self._end - self._begin > 0:
            left = self._buffer[self._begin - self._end:]
            self._buffer = left + data
        else:
            self._buffer = data

        self._begin = self._buffer
        self._end = self._begin + len(self._buffer)
        return 0


cdef class Decoder:
    def __init__(self, stream):
        self._decoder = CDecoder(stream)

    def __len__(self):
        return self._decoder.position()

    def read_field_number(self):
        return self._decoder.read_field_number()

    def read_field_number_and_wire_type(self):
        return self._decoder.read_field_number_and_wire_type()

    def read_sint32(self):
        return self._decoder.read_sint32()

    def read_uint32(self):
        return self._decoder.read_uint32()

    def read_sint64(self):
        return self._decoder.read_sint64()

    def read_uint64(self):
        return self._decoder.read_uint64()

    def read_bool(self):
        return self._decoder.read_bool()

    def read_double(self):
        return self._decoder.read_double()

    def read_float(self):
        return self._decoder.read_float()

    def read_string(self):
        return self._decoder.read_string()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import options
from ..compat import six, urlparse
from ..models import Projects
from ..rest import RestClient
from .errors import TunnelError

TUNNEL_VERSION = 5

_endpoint_cache = dict()


class BaseTunnel(object):
    def __init__(self, odps=None, client=None, project=None, endpoint=None, quota_name=None):
        self._client = odps.rest if odps is not None else client
        self._account = self._client.account
        if project is None and odps is None:
            raise AttributeError('%s requires project parameter.' % type(self).__name__)
        if isinstance(project, six.string_types):
            if odps is not None:
                self._project = odps.get_project(project or odps.project)
            else:
                self._project = Projects(client=self._client)[project]
        elif project is None:
            self._project = odps.get_project()
        else:
            self._project = project

        self._quota_name = quota_name or options.tunnel.quota_name
        if quota_name is not None:
            self._endpoint = endpoint
        else:
            self._endpoint = endpoint or self._project._tunnel_endpoint or options.tunnel.endpoint
        self._tunnel_rest = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def quota_name(self):
        return self._quota_name

    def _get_tunnel_server(self, project):
        protocol = urlparse(self._client.endpoint).scheme
        if protocol is None or protocol not in ('http', 'https'):
            raise TunnelError("Invalid protocol: %s" % protocol)

        ep_cache_key = (self._client.endpoint, project.name, self._quota_name)
        if ep_cache_key in _endpoint_cache:
            return _endpoint_cache[ep_cache_key]

        url = '/'.join([project.resource().rstrip('/'), 'tunnel'])
        params = {}
        if self._quota_name:
            params["quotaName"] = self._quota_name
        resp = self._client.get(url, action='service', params=params)

        if self._client.is_ok(resp):
            addr = resp.text
            server_ep = _endpoint_cache[ep_cache_key] = urlparse(
                '%s://%s' % (protocol, addr)
            ).geturl()
            return server_ep
        else:
            raise TunnelError("Can't get tunnel server address")

    @property
    def tunnel_rest(self):
        if self._tunnel_rest is not None:
            return self._tunnel_rest

        kw = dict(tag="TUNNEL")
        if options.data_proxy is not None:
            kw['proxy'] = options.data_proxy
        if getattr(self._client, 'app_account', None) is not None:
            kw['app_account'] = self._client.app_account

        endpoint = self._endpoint
        if endpoint is None:
            endpoint = self._get_tunnel_server(self._project)
        self._tunnel_rest = RestClient(self._account, endpoint, self._client.project, **kw)
        return self._tunnel_rest



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import time

from mars.scheduler.session import SessionManagerActor

from ....config import options

logger = logging.getLogger(__name__)

DEFAULT_MARS_IDLE_TIMEOUT = 3 * 3600
CUPID_LAST_IDLE_TIME_KEY = "MarsServiceLastIdleTime"


class CupidSessionManagerActor(SessionManagerActor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if "MARS_INSTANCE_IDLE_TIMEOUT" in os.environ:
            self._idle_timeout = int(os.environ["MARS_INSTANCE_IDLE_TIMEOUT"])
        else:
            self._idle_timeout = DEFAULT_MARS_IDLE_TIMEOUT

        self._last_active_time = time.time()
        self._last_active_time_from_service = None
        self._check_started = False

    def create_session(self, *args, **kwargs):
        session_ref = super().create_session(*args, **kwargs)
        self._last_activity_time = time.time()
        if not self._check_started and self._idle_timeout is not None:
            from cupid.runtime import context

            self._check_started = True
            kv_store = context().kv_store()
            last_idle_time_str = kv_store.get(CUPID_LAST_IDLE_TIME_KEY) or 0
            _, self._last_active_time_from_service = self._get_service_activity_info()
            if not last_idle_time_str:
                self._last_active_time = self._last_active_time_from_service
            else:
                self._last_active_time = float(last_idle_time_str)

            self.ref().check_instance_idle(_delay=10, _tell=True, _wait=False)
            logger.info(
                "Instance will go timeout in %s seconds when no active sessions.",
                self._idle_timeout,
            )
        return session_ref

    def _get_service_activity_info(self):
        last_active_time = self._last_active_time
        has_running = False
        for ref in self._session_refs.values():
            for info in ref.get_graph_infos().values():
                if info.get("end_time") is None:
                    has_running = True
                    break
                else:
                    last_active_time = max(info["end_time"], last_active_time)
            if has_running:
                break
        return has_running, last_active_time

    def check_instance_idle(self):
        from cupid.runtime import context

        has_running, active_time_from_service = self._get_service_activity_info()
        if active_time_from_service != self._last_active_time_from_service:
            self._last_active_time = active_time_from_service
            self._last_active_time_from_service = active_time_from_service
        elif has_running:
            self._last_active_time = time.time()

        if self._last_active_time < time.time() - self._idle_timeout:
            # timeout: we need to kill the instance
            logger.warning("Timeout met, killing the instance now.")
            self._stop_instance()
        else:
            kv_store = context().kv_store()
            kv_store[CUPID_LAST_IDLE_TIME_KEY] = str(self._last_active_time)
            self.ref().check_instance_idle(_delay=10, _tell=True, _wait=False)

    @staticmethod
    def _stop_instance():
        from cupid import context, ContainerStatus, WorkItemProgress
        from odps import ODPS
        from odps.accounts import BearerTokenAccount

        cupid_context = context()
        cupid_context.report_container_status(
            ContainerStatus.TERMINATED,
            "Instance idle timed out, stopping",
            WorkItemProgress.WIP_TERMINATING,
        )
        time.sleep(options.mars.container_status_timeout)

        # when instance is still not stopped, we kill forcifully
        bearer_token = cupid_context.get_bearer_token()
        account = BearerTokenAccount(bearer_token)
        project = os.environ["ODPS_PROJECT_NAME"]
        endpoint = os.environ["ODPS_RUNTIME_ENDPOINT"]
        o = ODPS(None, None, account=account, project=project, endpoint=endpoint)

        o.stop_instance(os.environ["MARS_K8S_POD_NAMESPACE"])


try:
    from ...internal.core import DEFAULT_MARS_IDLE_TIMEOUT
except ImportError:
    pass



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *

from . import errno

cdef:
    int BD_SUCCESS = errno.BD_SUCCESS
    int BD_BUFFER_EXHAUSTED = errno.BD_BUFFER_EXHAUSTED
    int BD_CHECKSUM_INVALID = errno.BD_CHECKSUM_INVALID
    int BD_COUNT_NOT_MATCH = errno.BD_COUNT_NOT_MATCH
    int BD_INVALID_STREAM_DATA = errno.BD_INVALID_STREAM_DATA
    int BD_INVALID_PB_TAG = errno.BD_INVALID_PB_TAG


cdef class Encoder:
    cdef init(self, char *buf_ptr, int buf_len):
        self._buf_ptr = buf_ptr
        self._buf_len = buf_len
        self._last_error = BD_SUCCESS
        self._pos = 0

    cdef int position(self) nogil:
        return self._pos

    cdef int get_last_error(self) nogil:
        return self._last_error

    cdef void set_last_error(self, int errno) nogil:
        self._last_error = errno

    cdef int append_tag(self, int field_num, int wire_type) nogil:
        cdef int key
        key = (field_num << 3) | wire_type
        cdef int size = self._set_varint64(key)
        return size

    cdef int append_sint32(self, int32_t value) nogil:
        return self._set_signed_varint32(value)

    cdef int append_uint32(self, uint32_t value) nogil:
        return self._set_varint32(value)

    cdef int append_sint64(self, int64_t value) nogil:
        return self._set_signed_varint64(value)

    cdef int append_uint64(self, uint64_t value) nogil:
        return self._set_varint64(value)

    cdef int append_bool(self, bint value) nogil:
        return self._set_varint32(value)

    cdef int append_float(self, float value) nogil:
        if self._pos + sizeof(float) >= self._buf_len:
            self._last_error = BD_BUFFER_EXHAUSTED
            return -1
        memcpy(self._buf_ptr + self._pos, &value, sizeof(float))
        self._pos += sizeof(float)
        return sizeof(float)

    cdef int append_double(self, double value) nogil:
        if self._pos + sizeof(double) >= self._buf_len:
            self._last_error = BD_BUFFER_EXHAUSTED
            return -1
        memcpy(self._buf_ptr + self._pos, &value, sizeof(double))
        self._pos += sizeof(double)
        return sizeof(double)

    cdef int append_string(self, const char *ptr, int value_len) nogil:
        cdef int size = self._set_varint32(value_len)
        if self._last_error:
            return -1
        if self._pos + value_len >= self._buf_len:
            self._last_error = BD_BUFFER_EXHAUSTED
            return -1
        memcpy(self._buf_ptr + self._pos, ptr, value_len)
        return size + value_len

    cdef int _set_varint32(self, int32_t varint) nogil:
        """
        Serialize an integer into a protobuf varint; return the number of bytes
        serialized.
        """

        # Negative numbers are always 10 bytes, so we need a uint64_t to
        # facilitate encoding
        cdef uint64_t enc = varint
        bits = enc & 0x7f
        enc >>= 7
        cdef int idx = 0
        while enc:
            if self._pos + idx >= self._buf_len:
                self._last_error = BD_BUFFER_EXHAUSTED
                return -1
            else:
                self._buf_ptr[self._pos + idx] = bits | 0x80
            bits = enc & 0x7f
            enc >>= 7
            idx += 1

        if self._pos + idx >= self._buf_len:
            self._last_error = BD_BUFFER_EXHAUSTED
            return -1
        self._buf_ptr[self._pos + idx] = bits
        self._pos += idx + 1
        return idx + 1

    cdef int _set_varint64(self, int64_t varint) nogil:
        """
        Serialize an integer into a protobuf varint; return the number of bytes
        serialized.
        """

        # Negative numbers are always 10 bytes, so we need a uint64_t to
        # facilitate encoding
        cdef uint64_t enc = varint
        bits = enc & 0x7f
        enc >>= 7
        cdef int idx = 0
        while enc:
            if self._pos + idx >= self._buf_len:
                self._last_error = BD_BUFFER_EXHAUSTED
                return -1
            else:
                self._buf_ptr[self._pos + idx] = bits | 0x80
            bits = enc & 0x7f
            enc >>= 7
            idx += 1

        if self._pos + idx >= self._buf_len:
            self._last_error = BD_BUFFER_EXHAUSTED
            return -1
        self._buf_ptr[self._pos + idx] = bits
        self._pos += idx + 1
        return idx + 1

    cdef int _set_signed_varint32(self, int32_t varint) nogil:
        """
        Serialize an integer into a signed protobuf varint; return the number of
        bytes serialized.
        """
        cdef uint32_t enc
        cdef int idx = 0

        enc = (varint << 1) ^ (varint >> 31) # zigzag encoding
        bits = enc & 0x7f
        enc >>= 7
        while enc:
            if self._pos + idx >= self._buf_len:
                self._last_error = BD_BUFFER_EXHAUSTED
                return -1
            else:
                self._buf_ptr[self._pos + idx] = bits | 0x80
            bits = enc & 0x7f
            enc >>= 7
            idx += 1

        if self._pos + idx >= self._buf_len:
            self._last_error = BD_BUFFER_EXHAUSTED
            return -1
        self._buf_ptr[self._pos + idx] = bits
        self._pos += idx + 1
        return idx + 1


    cdef int _set_signed_varint64(self, int64_t varint) nogil:
        """
        Serialize an integer into a signed protobuf varint; return the number of
        bytes serialized.
        """
        cdef uint64_t enc
        cdef int idx = 0

        enc = (varint << 1) ^ (varint >> 63) # zigzag encoding
        bits = enc & 0x7f
        enc >>= 7
        while enc:
            if self._pos + idx >= self._buf_len:
                self._last_error = BD_BUFFER_EXHAUSTED
                return -1
            else:
                self._buf_ptr[self._pos + idx] = bits | 0x80
            bits = enc & 0x7f
            enc >>= 7
            idx += 1

        if self._pos + idx >= self._buf_len:
            self._last_error = BD_BUFFER_EXHAUSTED
            return -1
        self._buf_ptr[self._pos + idx] = bits
        self._pos += idx + 1
        return idx + 1



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *
from libcpp.vector cimport vector

from ...src.types_c cimport SchemaSnapshot
from ...src.utils_c cimport CMillisecondsConverter
from ..pb.decoder_c cimport CDecoder
from ..checksum_c cimport Checksum


ctypedef int (*_SET_FUNCTION)(BaseTunnelRecordReader self, list record, int i) except? -1


cdef class BaseTunnelRecordReader:
    cdef public object _schema
    cdef object _columns
    cdef CMillisecondsConverter _mills_converter
    cdef CMillisecondsConverter _mills_converter_utc
    cdef object _to_date
    cdef CDecoder _reader
    cdef Checksum _crc
    cdef Checksum _crccrc
    cdef int _curr_cursor
    cdef int _n_columns
    cdef int _read_limit
    cdef bint _overflow_date_as_none
    cdef bint _struct_as_dict
    cdef vector[_SET_FUNCTION] _column_setters
    cdef SchemaSnapshot _schema_snapshot
    cdef list _partition_vals

    cdef object _read_struct(self, object value_type)
    cdef object _read_element(self, int data_type_id, object data_type)
    cdef list _read_array(self, object value_type)
    cdef bytes _read_string(self)
    cdef float _read_float(self) except? -1.0 nogil
    cdef double _read_double(self) except? -1.0 nogil
    cdef bint _read_bool(self) except? False nogil
    cdef int64_t _read_bigint(self) except? -1 nogil
    cdef object _read_datetime(self)
    cdef object _read_date(self)
    cdef object _read_timestamp_base(self, bint ntz)
    cdef object _read_timestamp(self)
    cdef object _read_timestamp_ntz(self)
    cdef object _read_interval_day_time(self)
    cdef int _set_record_list_value(self, list record, int i, object value) except? -1
    cdef int _set_string(self, list record, int i) except? -1
    cdef int _set_float(self, list record, int i) except? -1
    cdef int _set_double(self, list record, int i) except? -1
    cdef int _set_bool(self, list record, int i) except? -1
    cdef int _set_bigint(self, list record, int i) except? -1
    cdef int _set_datetime(self, list record, int i) except? -1
    cdef int _set_date(self, list record, int i) except? -1
    cdef int _set_decimal(self, list record, int i) except? -1
    cdef int _set_timestamp(self, list record, int i) except? -1
    cdef int _set_timestamp_ntz(self, list record, int i) except? -1
    cdef int _set_interval_day_time(self, list record, int i) except? -1
    cdef int _set_interval_year_month(self, list record, int i) except? -1
    cdef int _set_json(self, list record, int i) except? -1
    cpdef read(self)



import re
import sys
from types import CodeType
from types import FrameType
from types import TracebackType

try:
    from __pypy__ import tproxy
except ImportError:
    tproxy = None
try:
    from .cpython import tb_set_next
except ImportError:
    tb_set_next = None

if not tb_set_next and not tproxy:
    raise ImportError("Cannot use tblib. Runtime not supported.")

__version__ = '1.7.0'
__all__ = 'Traceback', 'TracebackParseError', 'Frame', 'Code'

PY3 = sys.version_info[0] == 3
FRAME_RE = re.compile(r'^\s*File "(?P<co_filename>.+)", line (?P<tb_lineno>\d+)(, in (?P<co_name>.+))?$')


class _AttrDict(dict):
    __slots__ = ()

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)


# noinspection PyPep8Naming
class __traceback_maker(Exception):
    pass


class TracebackParseError(Exception):
    pass


class Code(object):
    """
    Class that replicates just enough of the builtin Code object to enable serialization and traceback rendering.
    """
    co_code = None

    def __init__(self, code):
        self.co_filename = code.co_filename
        self.co_name = code.co_name
        self.co_argcount = 0
        self.co_kwonlyargcount = 0
        self.co_varnames = ()
        self.co_nlocals = 0
        self.co_stacksize = 0
        self.co_flags = 64
        self.co_firstlineno = 0

    # noinspection SpellCheckingInspection
    def __tproxy__(self, operation, *args, **kwargs):
        """
        Necessary for PyPy's tproxy.
        """
        if operation in ('__getattribute__', '__getattr__'):
            return getattr(self, args[0])
        else:
            return getattr(self, operation)(*args, **kwargs)


class Frame(object):
    """
    Class that replicates just enough of the builtin Frame object to enable serialization and traceback rendering.
    """
    def __init__(self, frame):
        self.f_locals = {}
        self.f_globals = dict(
            (k, v)
            for k, v in frame.f_globals.items()
            if k in ("__file__", "__name__")
        )
        self.f_code = Code(frame.f_code)
        self.f_lineno = frame.f_lineno

    def clear(self):
        """
        For compatibility with PyPy 3.5;
        clear() was added to frame in Python 3.4
        and is called by traceback.clear_frames(), which
        in turn is called by unittest.TestCase.assertRaises
        """

    # noinspection SpellCheckingInspection
    def __tproxy__(self, operation, *args, **kwargs):
        """
        Necessary for PyPy's tproxy.
        """
        if operation in ('__getattribute__', '__getattr__'):
            if args[0] == 'f_code':
                return tproxy(CodeType, self.f_code.__tproxy__)
            else:
                return getattr(self, args[0])
        else:
            return getattr(self, operation)(*args, **kwargs)


class Traceback(object):
    """
    Class that wraps builtin Traceback objects.
    """
    tb_next = None

    def __init__(self, tb):
        self.tb_frame = Frame(tb.tb_frame)
        # noinspection SpellCheckingInspection
        self.tb_lineno = int(tb.tb_lineno)

        # Build in place to avoid exceeding the recursion limit
        tb = tb.tb_next
        prev_traceback = self
        cls = type(self)
        while tb is not None:
            traceback = object.__new__(cls)
            traceback.tb_frame = Frame(tb.tb_frame)
            traceback.tb_lineno = int(tb.tb_lineno)
            prev_traceback.tb_next = traceback
            prev_traceback = traceback
            tb = tb.tb_next

    def as_traceback(self):
        """
        Convert to a builtin Traceback object that is usable for raising or rendering a stacktrace.
        """
        if tproxy:
            return tproxy(TracebackType, self.__tproxy__)
        if not tb_set_next:
            raise RuntimeError("Unsupported Python interpreter!")

        current = self
        top_tb = None
        tb = None
        while current:
            f_code = current.tb_frame.f_code
            code = compile('\n' * (current.tb_lineno - 1) + 'raise __traceback_maker', current.tb_frame.f_code.co_filename, 'exec')
            if hasattr(code, "replace"):
                # Python 3.8 and newer
                code = code.replace(co_argcount=0,
                                    co_filename=f_code.co_filename, co_name=f_code.co_name,
                                    co_freevars=(), co_cellvars=())
            elif PY3:
                code = CodeType(
                    0, code.co_kwonlyargcount,
                    code.co_nlocals, code.co_stacksize, code.co_flags,
                    code.co_code, code.co_consts, code.co_names, code.co_varnames,
                    f_code.co_filename, f_code.co_name,
                    code.co_firstlineno, code.co_lnotab, (), ()
                )
            else:
                code = CodeType(
                    0,
                    code.co_nlocals, code.co_stacksize, code.co_flags,
                    code.co_code, code.co_consts, code.co_names, code.co_varnames,
                    f_code.co_filename.encode(), f_code.co_name.encode(),
                    code.co_firstlineno, code.co_lnotab, (), ()
                )

            # noinspection PyBroadException
            try:
                exec(code, dict(current.tb_frame.f_globals), {})
            except Exception:
                next_tb = sys.exc_info()[2].tb_next
                if top_tb is None:
                    top_tb = next_tb
                if tb is not None:
                    tb_set_next(tb, next_tb)
                tb = next_tb
                del next_tb

            current = current.tb_next
        try:
            return top_tb
        finally:
            del top_tb
            del tb
    to_traceback = as_traceback

    # noinspection SpellCheckingInspection
    def __tproxy__(self, operation, *args, **kwargs):
        """
        Necessary for PyPy's tproxy.
        """
        if operation in ('__getattribute__', '__getattr__'):
            if args[0] == 'tb_next':
                return self.tb_next and self.tb_next.as_traceback()
            elif args[0] == 'tb_frame':
                return tproxy(FrameType, self.tb_frame.__tproxy__)
            else:
                return getattr(self, args[0])
        else:
            return getattr(self, operation)(*args, **kwargs)

    def as_dict(self):
        """
        Converts to a dictionary representation. You can serialize the result to JSON as it only has
        builtin objects like dicts, lists, ints or strings.
        """
        if self.tb_next is None:
            tb_next = None
        else:
            tb_next = self.tb_next.to_dict()

        code = {
            'co_filename': self.tb_frame.f_code.co_filename,
            'co_name': self.tb_frame.f_code.co_name,
        }
        frame = {
            'f_globals': self.tb_frame.f_globals,
            'f_code': code,
            'f_lineno': self.tb_frame.f_lineno,
        }
        return {
            'tb_frame': frame,
            'tb_lineno': self.tb_lineno,
            'tb_next': tb_next,
        }
    to_dict = as_dict

    @classmethod
    def from_dict(cls, dct):
        """
        Creates an instance from a dictionary with the same structure as ``.as_dict()`` returns.
        """
        if dct['tb_next']:
            tb_next = cls.from_dict(dct['tb_next'])
        else:
            tb_next = None

        code = _AttrDict(
            co_filename=dct['tb_frame']['f_code']['co_filename'],
            co_name=dct['tb_frame']['f_code']['co_name'],
        )
        frame = _AttrDict(
            f_globals=dct['tb_frame']['f_globals'],
            f_code=code,
            f_lineno=dct['tb_frame']['f_lineno'],
        )
        tb = _AttrDict(
            tb_frame=frame,
            tb_lineno=dct['tb_lineno'],
            tb_next=tb_next,
        )
        return cls(tb)

    @classmethod
    def from_string(cls, string, strict=True):
        """
        Creates an instance by parsing a stacktrace. Strict means that parsing stops when lines are not indented by at least two spaces
        anymore.
        """
        frames = []
        header = strict

        for line in string.splitlines():
            line = line.rstrip()
            if header:
                if line == 'Traceback (most recent call last):':
                    header = False
                continue
            frame_match = FRAME_RE.match(line)
            if frame_match:
                frames.append(frame_match.groupdict())
            elif line.startswith('  '):
                pass
            elif strict:
                break  # traceback ended

        if frames:
            previous = None
            for frame in reversed(frames):
                previous = _AttrDict(
                    frame,
                    tb_frame=_AttrDict(
                        frame,
                        f_globals=_AttrDict(
                            __file__=frame['co_filename'],
                            __name__='?',
                        ),
                        f_code=_AttrDict(frame),
                        f_lineno=int(frame['tb_lineno']),
                    ),
                    tb_next=previous,
                )
            return cls(previous)
        else:
            raise TracebackParseError("Could not find any frames in %r." % string)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
import copy
import glob
import hashlib
import json
import logging
import os
import platform
import stat
import subprocess
import sys
import tempfile
import threading
import time
import uuid

from .accounts import AliyunAccount
from .compat import pickle, six, builtins, futures
from .config import options
from .errors import NoSuchObject
from . import utils

TEMP_ROOT = utils.build_pyodps_dir('tempobjs')
SESSION_KEY = '%d_%s' % (int(time.time()), uuid.uuid4())
CLEANER_THREADS = 100
USER_FILE_RIGHTS = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR

CLEANUP_SCRIPT_TMPL = u"""
#-*- coding:utf-8 -*-
import os
import sys
import json

try:
    os.unlink(os.path.realpath(__file__))
except Exception:
    pass

temp_codes = json.loads({odps_info!r})
import_paths = json.loads({import_paths!r})
biz_ids = json.loads({biz_ids!r})

if sys.version_info[0] < 3:
    if sys.platform == 'win32':
        import_paths = [p.encode('mbcs') for p in import_paths]
    else:
        import_paths = [p.encode() for p in import_paths]

normed_paths = set(os.path.normcase(os.path.normpath(p)) for p in sys.path)
import_paths = [p for p in import_paths
                if os.path.normcase(os.path.normpath(p)) not in normed_paths]

sys.path.extend(import_paths)
from odps import ODPS, tempobj

if os.environ.get('WAIT_CLEANUP') == '1':
    tempobj.cleanup_timeout = None
else:
    tempobj.cleanup_timeout = 5
tempobj.cleanup_mode = True
tempobj.host_pid = {host_pid}
tempobj.ObjectRepositoryLib.biz_ids = set(biz_ids)

for o_desc in temp_codes:
    ODPS(**o_desc)
os._exit(0)
""".lstrip()


cleanup_mode = False
cleanup_timeout = 0
host_pid = os.getpid()
if six.PY3:  # make flake8 happy
    unicode = str


class ExecutionEnv(object):
    def __init__(self, **kwargs):
        self.cleaned = False
        self.os = os
        self.sys = sys
        self._g_env = copy.copy(globals())
        self.is_windows = 'windows' in platform.platform().lower()
        self.pid = os.getpid()
        self.os_sep = os.sep
        self.executable = sys.executable
        self.six = six

        import_paths = copy.deepcopy(sys.path)
        package_root = os.path.dirname(__file__)
        if package_root not in import_paths:
            import_paths.append(package_root)
        self.import_path_json = utils.to_text(json.dumps(import_paths, ensure_ascii=False))

        self.builtins = builtins
        self.io = __import__('io', fromlist=[''])
        if six.PY3:
            self.conv_bytes = (lambda s: s.encode() if isinstance(s, str) else s)
            self.conv_unicode = (lambda s: s if isinstance(s, str) else s.decode())
        else:
            self.conv_bytes = (lambda s: s.encode() if isinstance(s, unicode) else s)
            self.conv_unicode = (lambda s: s if isinstance(s, unicode) else s.decode())
        self.subprocess = subprocess
        self.temp_dir = tempfile.gettempdir()
        self.template = CLEANUP_SCRIPT_TMPL
        self.file_right = USER_FILE_RIGHTS
        self.is_main_process = utils.is_main_process()
        for k, v in six.iteritems(kwargs):
            setattr(self, k, v)


class TempObject(object):
    __slots__ = ()
    _type = ''
    _priority = 0

    def __init__(self, *args, **kwargs):
        for k, v in zip(self.__slots__, args):
            setattr(self, k, v)
        for k in self.__slots__:
            if hasattr(self, k):
                continue
            setattr(self, k, kwargs.get(k))

    def __hash__(self):
        if self.__slots__:
            return hash(tuple(getattr(self, k, None) for k in self.__slots__))
        return super(TempObject, self).__hash__()

    def __eq__(self, other):
        if not isinstance(other, TempObject):
            return False
        if self._type != other._type:
            return False
        return all(getattr(self, k) == getattr(other, k) for k in self.__slots__)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getstate__(self):
        return {slot: getattr(self, slot) for slot in self.__slots__ if hasattr(self, slot)}

    def __setstate__(self, state):
        for slot, value in state.items():
            setattr(self, slot, value)


class TempTable(TempObject):
    __slots__ = 'table', 'project', 'schema'
    _type = 'Table'

    def drop(self, odps):
        odps.delete_table(self.table, project=self.project, schema=self.schema, async_=True)


class TempModel(TempObject):
    __slots__ = 'model', 'project', 'schema'
    _type = 'OfflineModel'

    def drop(self, odps):
        try:
            odps.delete_offline_model(self.model, project=self.project, schema=self.schema)
        except NoSuchObject:
            pass


class TempFunction(TempObject):
    __slots__ = 'function', 'project', 'schema'
    _type = 'Function'
    _priority = 1

    def drop(self, odps):
        try:
            odps.delete_function(self.function, project=self.project, schema=self.schema)
        except NoSuchObject:
            pass


class TempResource(TempObject):
    __slots__ = 'resource', 'project', 'schema'
    _type = 'Resource'

    def drop(self, odps):
        try:
            odps.delete_resource(self.resource, project=self.project, schema=self.schema)
        except NoSuchObject:
            pass


class TempVolumePartition(TempObject):
    __slots__ = 'volume', 'partition', 'project', 'schema'
    _type = 'VolumePartition'

    def drop(self, odps):
        try:
            odps.delete_volume_partition(
                self.volume, self.partition, project=self.project, schema=self.schema
            )
        except NoSuchObject:
            pass


class ObjectRepository(object):
    def __init__(self, file_name):
        self._container = set()
        self._file_name = file_name
        if file_name and os.path.exists(file_name):
            self.load()

    def put(self, obj, dump=True):
        self._container.add(obj)
        if dump:
            self.dump()

    def cleanup(self, odps, use_threads=True):
        cleaned = []

        def _cleaner(obj):
            try:
                obj.drop(odps)
                cleaned.append(obj)
            except:
                pass

        if self._container:
            if use_threads:
                pool = futures.ThreadPoolExecutor(CLEANER_THREADS)
                list(pool.map(_cleaner, reversed(list(self._container))))
            else:
                for o in sorted(list(self._container), key=lambda ro: type(ro)._priority, reverse=True):
                    _cleaner(o)
        for obj in cleaned:
            if obj in self._container:
                self._container.remove(obj)
        if not self._container and self._file_name:
            try:
                os.unlink(self._file_name)
            except OSError:
                pass
        else:
            self.dump()

    def dump(self):
        if self._file_name is None:
            return
        try:
            with open(self._file_name, 'wb') as outf:
                pickle.dump(list(self._container), outf, protocol=0)
                outf.close()
        except OSError:
            return
        os.chmod(self._file_name, USER_FILE_RIGHTS)

    def load(self):
        try:
            with open(self._file_name, 'rb') as inpf:
                contents = pickle.load(inpf)
            self._container.update(contents)
        except (EOFError, OSError):
            pass


class ObjectRepositoryLib(dict):
    biz_ids = set([options.biz_id, ]) if options.biz_id else set(['default', ])
    odps_info = dict()

    biz_ids_json = json.dumps(list(biz_ids))
    odps_info_json = json.dumps([v for v in six.itervalues(odps_info)])

    def __init__(self, *args, **kwargs):
        super(ObjectRepositoryLib, self).__init__(*args, **kwargs)
        self._env = ExecutionEnv()

    def __del__(self):
        self._exec_cleanup_script()

    @classmethod
    def add_biz_id(cls, biz_id):
        cls.biz_ids.add(biz_id)
        cls.biz_ids_json = json.dumps(list(cls.biz_ids))

    @classmethod
    def add_odps_info(cls, odps):
        odps_key = _gen_repository_key(odps)
        if odps_key is None:
            return

        cls.odps_info[odps_key] = dict(
            access_id=utils.to_str(odps.account.access_id),
            secret_access_key=utils.to_str(odps.account.secret_access_key),
            project=utils.to_str(odps.project),
            endpoint=utils.to_str(odps.endpoint),
        )
        cls.odps_info_json = json.dumps([v for v in six.itervalues(cls.odps_info)])

    def _exec_cleanup_script(self):
        global cleanup_mode

        if not self:
            return

        env = self._env
        if cleanup_mode or not env.is_main_process or env.cleaned:
            return
        env.cleaned = True

        script = env.template.format(import_paths=env.import_path_json, odps_info=self.odps_info_json,
                                     host_pid=env.pid, biz_ids=self.biz_ids_json)

        script_name = env.temp_dir + env.os_sep + 'tmp_' + str(env.pid) + '_cleanup_script.py'
        script_file = env.io.FileIO(script_name, 'w')
        script_file.write(env.conv_bytes(script))
        script_file.close()
        try:
            if env.is_windows:
                env.os.chmod(script_name, env.file_right)
            else:
                env.subprocess.call(['chmod', oct(env.file_right).replace('o', ''), script_name])
        except:
            pass

        kwargs = dict(close_fds=True)
        if env.is_windows:
            si = subprocess.STARTUPINFO()
            si.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            kwargs['startupinfo'] = si
        env.subprocess.call([env.executable, script_name], **kwargs)


_cleaned_keys = set()
_obj_repos = ObjectRepositoryLib()  # this line should be put last due to initialization dependency
atexit.register(_obj_repos._exec_cleanup_script)


def _is_pid_running(pid):
    if 'windows' in platform.platform().lower():
        task_lines = os.popen('TASKLIST /FI "PID eq {0}" /NH'.format(pid)).read().strip().splitlines()
        if not task_lines:
            return False
        return str(pid) in set(task_lines[0].split())
    else:
        try:
            os.kill(pid, 0)
            return True
        except OSError:
            return False


def clean_objects(odps, biz_ids=None, use_threads=False):
    odps_key = _gen_repository_key(odps)
    if odps_key is None:
        return

    files = []
    biz_ids = biz_ids or _obj_repos.biz_ids
    for biz_id in biz_ids:
        files.extend(glob.glob(os.path.join(TEMP_ROOT, biz_id, odps_key, '*.his')))

    for fn in files:
        repo = ObjectRepository(fn)
        repo.cleanup(odps, use_threads=use_threads)


def clean_stored_objects(odps):
    global cleanup_timeout, host_pid

    if not utils.is_main_process():
        return

    odps_key = _gen_repository_key(odps)
    if odps_key is None or odps_key in _cleaned_keys:
        return
    _cleaned_keys.add(odps_key)

    files = []
    for biz_id in _obj_repos.biz_ids:
        files.extend(glob.glob(os.path.join(TEMP_ROOT, biz_id, odps_key, '*.his')))

    def clean_thread():
        for fn in files:
            writer_pid = int(fn.rsplit('__', 1)[-1].split('.', 1)[0])

            # we do not clean running process, unless its pid equals host_pid
            if writer_pid != host_pid and _is_pid_running(writer_pid):
                continue

            repo = ObjectRepository(fn)
            repo.cleanup(odps)

    thread_obj = threading.Thread(target=clean_thread)
    thread_obj.start()
    if cleanup_timeout == 0:
        return
    else:
        if cleanup_timeout is not None and cleanup_timeout < 0:
            cleanup_timeout = None
        thread_obj.join(cleanup_timeout)


def _gen_repository_key(odps):
    if getattr(odps.account, 'access_id', None):
        keys = [odps.account.access_id, odps.endpoint, str(odps.project)]
    elif getattr(odps.account, 'token', None):
        keys = [utils.to_str(odps.account.token), odps.endpoint, str(odps.project)]
    else:
        return
    return hashlib.md5(utils.to_binary('####'.join(keys))).hexdigest()


def _put_objects(odps, objs):
    odps_key = _gen_repository_key(odps)
    if odps_key is None:
        return

    biz_id = options.biz_id if options.biz_id else 'default'
    ObjectRepositoryLib.add_biz_id(biz_id)
    if odps_key not in _obj_repos:
        if isinstance(odps.account, AliyunAccount):
            ObjectRepositoryLib.add_odps_info(odps)
        file_dir = os.path.join(TEMP_ROOT, biz_id, odps_key)
        try:
            if not os.path.exists(file_dir):
                os.makedirs(file_dir)
        except OSError:
            pass
        file_name = os.path.join(file_dir, 'temp_objs_{0}__{1}.his'.format(SESSION_KEY, os.getpid()))
        _obj_repos[odps_key] = ObjectRepository(file_name)
    [_obj_repos[odps_key].put(o, False) for o in objs]
    _obj_repos[odps_key].dump()


def register_temp_table(odps, table, project=None, schema=None):
    if isinstance(table, six.string_types):
        table = [table, ]
    _put_objects(odps, [
        TempTable(t, project or odps.project, schema=schema or odps.schema) for t in table
    ])


def register_temp_model(odps, model, project=None, schema=None):
    if isinstance(model, six.string_types):
        model = [model, ]
    _put_objects(odps, [
        TempModel(m, project or odps.project, schema=schema or odps.schema) for m in model
    ])


def register_temp_resource(odps, resource, project=None, schema=None):
    if isinstance(resource, six.string_types):
        resource = [resource, ]
    _put_objects(odps, [
        TempResource(r, project if project else odps.project, schema=schema or odps.schema) for r in resource
    ])


def register_temp_function(odps, func, project=None, schema=None):
    if isinstance(func, six.string_types):
        func = [func, ]
    _put_objects(odps, [
        TempFunction(f, project if project else odps.project, schema=schema or odps.schema) for f in func
    ])


def register_temp_volume_partition(odps, volume_partition_tuple, project=None, schema=None):
    if isinstance(volume_partition_tuple, tuple):
        volume_partition_tuple = [volume_partition_tuple, ]
    _put_objects(odps, [
        TempVolumePartition(v, p, project if project else odps.project, schema=schema or odps.schema)
        for v, p in volume_partition_tuple
    ])



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libc.stdint cimport *
from libc.string cimport *

from ...src.types_c cimport BaseRecord, SchemaSnapshot
from ...src.utils_c cimport CMillisecondsConverter
from ..checksum_c cimport Checksum
from ..pb.encoder_c cimport CEncoder

cdef class ProtobufRecordWriter:

    cdef int DEFAULT_BUFFER_SIZE
    cdef CEncoder _encoder
    cdef object _output
    cdef int _buffer_size
    cdef int _n_total
    cdef int _n_columns
    cdef int _last_flush_time

    cpdef _re_init(self, output)
    cpdef flush(self)
    cpdef close(self)
    cpdef flush_all(self)
    cpdef int _refresh_buffer(self) except -1
    cdef int _write_tag(self, int field_num, int wire_type) except + nogil
    cdef int _write_raw_long(self, int64_t val) except + nogil
    cdef int _write_raw_int(self, int32_t val) except + nogil
    cdef int _write_raw_uint(self, uint32_t val) except + nogil
    cdef int _write_raw_bool(self, bint val) except + nogil
    cdef int _write_raw_float(self, float val) except + nogil
    cdef int _write_raw_double(self, double val) except + nogil
    cdef int _write_raw_string(self, const char *ptr, uint32_t size) except + nogil


cdef class BaseRecordWriter(ProtobufRecordWriter):
    cdef object _encoding
    cdef bint _is_utf8
    cdef object _schema
    cdef object _columns
    cdef size_t _curr_cursor_c
    cdef object _to_days
    cdef object _reader_schema
    cdef CMillisecondsConverter _mills_converter
    cdef CMillisecondsConverter _mills_converter_utc
    cdef Checksum _crc_c
    cdef Checksum _crccrc_c
    cdef SchemaSnapshot _schema_snapshot

    cpdef write(self, BaseRecord record)
    cdef void _write_bool(self, bint data) except + nogil
    cdef void _write_long(self, int64_t data) except + nogil
    cdef void _write_float(self, float data) except + nogil
    cdef void _write_double(self, double data) except + nogil
    cdef _write_string(self, object data)
    cdef _write_timestamp_base(self, object data, bint ntz)
    cdef _write_timestamp(self, object data)
    cdef _write_timestamp_ntz(self, object data)
    cdef _write_interval_day_time(self, object data)
    cdef _write_field(self, object val, int data_type_id, object data_type)
    cdef _write_array(self, object data, object data_type)
    cdef _write_struct(self, object data, object data_type)
    cpdef _write_finish_tags(self)
    cpdef close(self)


#!/usr/bin/env python
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Parts of this file were taken from the pandas project
# (https://github.com/pandas-dev/pandas), which is permitted for use under
# the BSD 3-Clause License

import os
import platform
import shutil
import sys

from setuptools import setup, find_packages, Extension
from setuptools.command.install import install

try:
    from setuptools import Command
except ImportError:
    from distutils.cmd import Command
try:
    from sysconfig import get_config_var
except ImportError:
    from distutils.sysconfig import get_config_var
try:
    from packaging.version import Version
except ImportError:
    from distutils.version import LooseVersion as Version

# From https://github.com/pandas-dev/pandas/pull/24274:
# For mac, ensure extensions are built for macos 10.9 when compiling on a
# 10.9 system or above, overriding distuitls behaviour which is to target
# the version that python was built for. This may be overridden by setting
# MACOSX_DEPLOYMENT_TARGET before calling setup.py
if sys.platform == 'darwin':
    if 'MACOSX_DEPLOYMENT_TARGET' not in os.environ:
        current_system = Version(platform.mac_ver()[0])
        python_target = Version(get_config_var('MACOSX_DEPLOYMENT_TARGET'))
        if python_target < Version('10.9') and current_system >= Version('10.9'):
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = '10.9'

repo_root = os.path.dirname(os.path.abspath(__file__))

try:
    execfile
except NameError:
    def execfile(fname, globs, locs=None):
        locs = locs or globs
        exec(compile(open(fname).read(), fname, "exec"), globs, locs)

version_ns = {}
execfile(os.path.join(repo_root, 'odps', '_version.py'), version_ns)

extra_install_cmds = []


def which(program):
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


# http://stackoverflow.com/questions/12683834/how-to-copy-directory-recursively-in-python-and-overwrite-all
def recursive_overwrite(src, dest, filter_func=None):
    destinations = []
    filter_func = filter_func or (lambda s: True)
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        files = os.listdir(src)
        for f in files:
            if not filter_func(f):
                continue
            destinations.extend(
                recursive_overwrite(os.path.join(src, f), os.path.join(dest, f))
            )
    else:
        shutil.copyfile(src, dest)
        destinations.append(dest)
    return destinations


class CustomInstall(install):
    def run(self):
        global extra_install_cmds
        install.run(self)
        [self.run_command(cmd) for cmd in extra_install_cmds]

version = sys.version_info
PY2 = version[0] == 2
PY3 = version[0] == 3
PYPY = platform.python_implementation().lower() == 'pypy'

if PY2 and version[:2] < (2, 7):
    raise Exception('PyODPS supports Python 2.7+ (including Python 3+).')

try:
    import distribute
    raise Exception("PyODPS cannot be installed when 'distribute' is installed. "
                    "Please uninstall it before installing PyODPS.")
except ImportError:
    pass

try:
    import pip
    for pk in pip.get_installed_distributions():
        if pk.key == 'odps':
            raise Exception('Package `odps` collides with PyODPS. Please uninstall it before installing PyODPS.')
except (ImportError, AttributeError):
    pass

try:
    from jupyter_core.paths import jupyter_data_dir
    has_jupyter = True
except ImportError:
    has_jupyter = False
try:
    from jupyterlab import __version__
    has_jupyterlab = True
except ImportError:
    has_jupyterlab = False

if len(sys.argv) > 1 and sys.argv[1] == 'clean':
    build_cmd = sys.argv[1]
else:
    build_cmd = None

requirements = []
with open('requirements.txt') as f:
    requirements.extend(f.read().splitlines())

full_requirements = [
    'jupyter>=1.0.0',
    'ipython>=4.0.0',
    'numpy>=1.6.0',
    'pandas>=0.17.0',
    'matplotlib>=1.4',
    'graphviz>=0.4',
    'greenlet>=0.4.10',
]
mars_requirements = [
    'pymars>=0.5.4',
    'protobuf>=3.6,<4.0',
]
if sys.version_info[0] == 2:
    full_requirements.append('ipython<6.0.0')
if sys.platform != 'win32':
    full_requirements.append('cython>=0.20')

long_description = None
if os.path.exists('README.rst'):
    with open('README.rst') as f:
        long_description = f.read()

setup_options = dict(
    name='pyodps',
    version=version_ns['__version__'],
    description='ODPS Python SDK and data analysis framework',
    long_description=long_description,
    author='Wu Wei',
    author_email='weiwu@cacheme.net',
    maintainer='Wenjun Si',
    maintainer_email='wenjun.swj@alibaba-inc.com',
    url='http://github.com/aliyun/aliyun-odps-python-sdk',
    license='Apache License 2.0',
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries',
    ],
    cmdclass={'install': CustomInstall},
    packages=find_packages(exclude=('*.tests.*', '*.tests')),
    include_package_data=True,
    install_requires=requirements,
    include_dirs=[],
    extras_require={'full': full_requirements, 'mars': mars_requirements},
    entry_points={
        'sqlalchemy.dialects': [
            'odps = odps.sqlalchemy_odps:ODPSDialect',
            'maxcompute = odps.sqlalchemy_odps:ODPSDialect',
        ],
        'superset.db_engine_specs': [
            'odps = odps.superset_odps:ODPSEngineSpec',
        ],
        'console_scripts': [
            'pyou = odps_scripts.pyou:main',
            'pyodps-pack = odps_scripts.pyodps_pack:main',
        ],
    },
)

if build_cmd != 'clean' and not PYPY:  # skip cython in pypy
    try:
        from Cython.Build import cythonize
        from Cython.Distutils import build_ext
        import cython

        # detect if cython works
        if sys.platform == 'win32':
            cython.inline('return a + b', a=1, b=1)

        cythonize_kw = dict(language_level=sys.version_info[0])
        extension_kw = dict(language='c++', include_dirs=[])
        if "MSC" in sys.version:
            extra_compile_args = ["/Ot", "/I" + os.path.join(repo_root, "misc")]
            extension_kw["extra_compile_args"] = extra_compile_args
        else:
            extra_compile_args = ["-O3"]
            extension_kw["extra_compile_args"] = extra_compile_args

        if os.environ.get("CYTHON_TRACE"):
            extension_kw["define_macros"] = [
                ("CYTHON_TRACE_NOGIL", "1"),
                ("CYTHON_TRACE", "1"),
            ]
            cythonize_kw["compiler_directives"] = {"linetrace": True}

        extensions = [
            Extension('odps.src.types_c', ['odps/src/types_c.pyx'], **extension_kw),
            Extension('odps.src.crc32c_c', ['odps/src/crc32c_c.pyx'], **extension_kw),
            Extension('odps.src.utils_c', ['odps/src/utils_c.pyx'], **extension_kw),
            Extension('odps.tunnel.pb.encoder_c', ['odps/tunnel/pb/encoder_c.pyx'], **extension_kw),
            Extension('odps.tunnel.pb.decoder_c', ['odps/tunnel/pb/decoder_c.pyx'], **extension_kw),
            Extension('odps.tunnel.io.writer_c', ['odps/tunnel/io/writer_c.pyx'], **extension_kw),
            Extension('odps.tunnel.io.reader_c', ['odps/tunnel/io/reader_c.pyx'], **extension_kw),
            Extension('odps.tunnel.checksum_c', ['odps/tunnel/checksum_c.pyx'], **extension_kw),
            Extension('odps.tunnel.hasher_c', ['odps/tunnel/hasher_c.pyx'], **extension_kw),
        ]
        try:
            import numpy as np
            np_extension_kw = extension_kw.copy()
            np_extension_kw['include_dirs'].append(np.get_include())
            extensions.extend([
                Extension('odps.tunnel.pdio.pdreader_c', ['odps/tunnel/pdio/pdreader_c.pyx'], **np_extension_kw),
                Extension('odps.tunnel.pdio.pdwriter_c', ['odps/tunnel/pdio/pdwriter_c.pyx'], **np_extension_kw),
                Extension('odps.tunnel.pdio.block_decoder_c', ['odps/tunnel/pdio/block_decoder_c.pyx'], **np_extension_kw),
                Extension('odps.tunnel.pdio.block_encoder_c', ['odps/tunnel/pdio/block_encoder_c.pyx'], **np_extension_kw),
            ])
            setup_options['include_dirs'].append(np.get_include())
        except ImportError:
            pass

        setup_options['cmdclass'].update({'build_ext': build_ext})
        force_recompile = bool(int(os.getenv("CYTHON_FORCE_RECOMPILE", "0")))
        setup_options['ext_modules'] = cythonize(extensions, force=force_recompile, **cythonize_kw)
    except:
        pass

if build_cmd != 'clean' and has_jupyter:
    class InstallJS(Command):
        description = "install JavaScript extensions"
        user_options = []

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            src_dir = os.path.join(repo_root, 'odps', 'static', 'ui', 'target')
            dest_dir = os.path.join(jupyter_data_dir(), 'nbextensions', 'pyodps')
            if os.path.exists(dest_dir):
                shutil.rmtree(dest_dir)
            if not os.path.exists(dest_dir):
                os.makedirs(dest_dir)
            recursive_overwrite(src_dir, dest_dir)

            try:
                from notebook.nbextensions import enable_nbextension
            except ImportError:
                return
            enable_nbextension('notebook', 'pyodps/main')


    class BuildJS(Command):
        description = "build JavaScript files"
        user_options = [
            ('registry=', None, 'npm registry')
        ]

        def initialize_options(self):
            self.registry = None

        def finalize_options(self):
            pass

        def run(self):
            if not which('npm'):
                raise Exception('You need to install npm before building the scripts.')

            cwd = os.getcwd()

            os.chdir(os.path.join(os.path.abspath(os.getcwd()), 'odps', 'static', 'ui'))
            cmd = 'npm install'
            if getattr(self, 'registry', None):
                cmd += ' --registry=' + self.registry
            print('executing ' + cmd)
            ret = os.system(cmd)
            ret >>= 8
            if ret != 0:
                print(cmd + ' exited with error: %d' % ret)

            print('executing grunt')
            ret = os.system('npm run grunt')
            ret >>= 8
            if ret != 0:
                print('grunt exited with error: %d' % ret)

            os.chdir(cwd)


    setup_options['cmdclass'].update({'install_js': InstallJS, 'build_js': BuildJS})
    extra_install_cmds.append('install_js')

if build_cmd != 'clean' and has_jupyterlab:
    class InstallJupyterLabExtension(Command):
        description = "install Jupyterlab Extension"
        user_options = [
            ('registry=', 'r', 'npm registry')
        ]

        def initialize_options(self):
            self.registry = 'https://registry.npm.taobao.org'

        def finalize_options(self):
            pass

        def run(self):
            lab_location = os.path.join(os.path.abspath(os.getcwd()), 'odps', 'lab_extension')
            if not os.path.exists(lab_location):
                return
            os.chdir(lab_location)
            print("\033[1;34m" + "Install pyodps-lab-extension" + "\033[0;0m")
            os.system('npm install --registry=' + self.registry)
            os.system('pip install .')
            print("\033[0;32m" + "pyodps-lab-extension install success" + "\033[0;0m")


    setup_options['cmdclass'].update({'install_jlab': InstallJupyterLabExtension})
    extra_install_cmds.append('install_jlab')

setup(**setup_options)

if build_cmd == 'clean':
    for root, dirs, files in os.walk(os.path.normpath('odps/')):
        pyx_files = set()
        c_file_pairs = []
        if '__pycache__' in dirs:
            full_path = os.path.join(root, '__pycache__')
            print("removing '%s'" % full_path)
            shutil.rmtree(full_path)
        for f in files:
            fn, ext = os.path.splitext(f)
            # delete compiled binaries
            if ext.lower() in ('.pyd', '.so', '.pyc'):
                full_path = os.path.join(root, f)
                print("removing '%s'" % full_path)
                os.unlink(full_path)
            elif ext.lower() == '.pyx':
                pyx_files.add(fn)
            elif ext.lower() in ('.c', '.cpp', '.cc'):
                c_file_pairs.append((fn, f))

        # remove cython-generated files
        for cfn, cf in c_file_pairs:
            if cfn in pyx_files:
                full_path = os.path.join(root, cf)
                print("removing '%s'" % full_path)
                os.unlink(full_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from datetime import datetime

from .. import serializers, types, utils
from .core import LazyLoad, XMLRemoteModel, JSONRemoteModel
from .storage_tier import StorageTierInfo


class Partition(LazyLoad):
    """
    A partition is a collection of rows in a table whose partition columns are equal to specific
    values.

    In order to write data into partition, users should call the ``open_writer``
    method with **with statement**. At the same time, the ``open_reader`` method is used
    to provide the ability to read records from a partition. The behavior of these
    methods are the same as those in Table class except that there are no 'partition' params.
    """
    __slots__ = 'spec', 'creation_time', 'last_meta_modified_time', \
        'last_data_modified_time', 'size', '_is_extend_info_loaded', \
        'is_archived', 'is_exstore', 'lifecycle', 'physical_size', \
        'file_num', 'reserved'

    class Column(XMLRemoteModel):

        name = serializers.XMLNodeAttributeField(attr='Name')
        value = serializers.XMLNodeAttributeField(attr='Value')

    class PartitionMeta(JSONRemoteModel):

        creation_time = serializers.JSONNodeField(
            'createTime', parse_callback=datetime.fromtimestamp, set_to_parent=True)
        last_meta_modified_time = serializers.JSONNodeField(
            'lastDDLTime', parse_callback=datetime.fromtimestamp, set_to_parent=True)
        last_data_modified_time = serializers.JSONNodeField(
            'lastModifiedTime', parse_callback=datetime.fromtimestamp, set_to_parent=True)
        size = serializers.JSONNodeField(
            'partitionSize', parse_callback=int, set_to_parent=True)

    class PartitionExtendedMeta(PartitionMeta):

        is_archived = serializers.JSONNodeField(
            'IsArchived', parse_callback=bool, set_to_parent=True)
        is_exstore = serializers.JSONNodeField(
            'IsExstore', parse_callback=bool, set_to_parent=True)
        lifecycle = serializers.JSONNodeField(
            'LifeCycle', parse_callback=int, set_to_parent=True)
        physical_size = serializers.JSONNodeField(
            'PhysicalSize', parse_callback=int, set_to_parent=True)
        file_num = serializers.JSONNodeField(
            'FileNum', parse_callback=int, set_to_parent=True)
        reserved = serializers.JSONNodeField(
            'Reserved', type='json', set_to_parent=True)

    columns = serializers.XMLNodesReferencesField(Column, 'Column')
    _schema = serializers.XMLNodeReferenceField(PartitionMeta, 'Schema')
    _extended_schema = serializers.XMLNodeReferenceField(PartitionExtendedMeta, 'Schema')

    def __init__(self, **kwargs):
        self._is_extend_info_loaded = False

        super(Partition, self).__init__(**kwargs)

    def __str__(self):
        return str(self.partition_spec)

    def __repr__(self):
        return '<Partition %s.`%s`(%s)>' % (
            str(self.table.project.name), str(self.table.name), str(self.partition_spec))

    def __getattribute__(self, attr):
        if attr in ('is_archived', 'is_exstore', 'lifecycle',
                    'physical_size', 'file_num', 'reserved'):
            if not self._is_extend_info_loaded:
                self.reload_extend_info()

            return object.__getattribute__(self, attr)

        val = object.__getattribute__(self, attr)
        if val is None and not self._loaded:
            if attr in getattr(Partition.PartitionMeta, '__fields'):
                self.reload()
                return object.__getattribute__(self, attr)

        return super(Partition, self).__getattribute__(attr)

    def _set_state(self, name, parent, client):
        self.__init__(spec=name, _parent=parent, _client=client)

    def _name(self):
        return

    @classmethod
    def get_partition_spec(cls, columns=None, spec=None):
        if spec is not None:
            return spec

        spec = types.PartitionSpec()
        for col in columns:
            spec[col.name] = col.value

        return spec

    @property
    def last_modified_time(self):
        warnings.warn(
            "Partition.last_modified_time is deprecated and will be replaced by "
            "Partition.last_data_modified_time.",
            DeprecationWarning,
            stacklevel=3,
        )
        utils.add_survey_call(".".join(
            [type(self).__module__, type(self).__name__, "last_modified_time"]
        ))
        return self.last_data_modified_time

    @property
    def partition_spec(self):
        return self.get_partition_spec(self._getattr('columns'), self._getattr('spec'))

    @property
    def name(self):
        return str(self.partition_spec)

    @property
    def table(self):
        return self.parent.parent

    @property
    def project(self):
        return self.table.project

    @property
    def storage_tier_info(self):
        return StorageTierInfo.deserial(self.reserved)

    def reload(self):
        url = self.resource()
        params = {'partition': str(self.partition_spec)}
        resp = self._client.get(url, params=params, curr_schema=self._get_schema_name())

        self.parse(self._client, resp, obj=self)

        self._loaded = True

    def reload_extend_info(self):
        url = self.resource()
        params = {'partition': str(self.partition_spec)}
        resp = self._client.get(
            url, action='extended', params=params, curr_schema=self._get_schema_name()
        )

        self.parse(self._client, resp, obj=self)
        self._is_extend_info_loaded = True

    def head(self, limit, columns=None):
        """
        Get the head records of a partition

        :param limit: records' size, 10000 at most
        :param list columns: the columns which is subset of the table columns
        :return: records
        :rtype: list

        .. seealso:: :class:`odps.models.Record`
        """
        return self.table.head(limit, partition=self.partition_spec, columns=columns)

    def to_df(self):
        """
        Create a PyODPS DataFrame from this partition.

        :return: DataFrame object
        """
        from ..df import DataFrame

        return DataFrame(self.table).filter_parts(self)

    @utils.with_wait_argument
    def drop(self, async_=False, if_exists=False):
        """
        Drop this partition.

        :param async_: run asynchronously if True
        :param if_exists:
        :return: None
        """
        return self.parent.delete(self, if_exists=if_exists, async_=async_)

    def open_reader(self, **kw):
        """
        Open the reader to read the entire records from this partition.

        :param reopen: the reader will reuse last one, reopen is true means open a new reader.
        :type reopen: bool
        :param endpoint: the tunnel service URL
        :param compress_option: compression algorithm, level and strategy
        :type compress_option: :class:`odps.tunnel.CompressOption`
        :param compress_algo: compression algorithm, work when ``compress_option`` is not provided,
                              can be ``zlib``, ``snappy``
        :param compress_level: used for ``zlib``, work when ``compress_option`` is not provided
        :param compress_strategy: used for ``zlib``, work when ``compress_option`` is not provided
        :return: reader, ``count`` means the full size, ``status`` means the tunnel status

        :Example:

        >>> with partition.open_reader() as reader:
        >>>     count = reader.count  # How many records of a partition
        >>>     for record in reader[0: count]:
        >>>         # read all data, actually better to split into reading for many times
        """
        return self.table.open_reader(str(self), **kw)

    def open_writer(self, blocks=None, **kw):
        return self.table.open_writer(self.partition_spec, blocks=blocks, **kw)

    @utils.with_wait_argument
    def truncate(self, async_=False):
        return self.table.truncate(self.partition_spec, async_=async_)

    @utils.with_wait_argument
    def set_storage_tier(self, storage_tier, async_=False):
        self._is_extend_info_loaded = False
        return self.table.set_storage_tier(
            storage_tier, partition_spec=self.partition_spec, async_=async_
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import warnings

from .compat import enum, six
from .config import options
from .core import ODPS
from .errors import NotSupportedError, InstanceTypeNotSupported, ODPSError
from .utils import to_str
from .models.session import PUBLIC_SESSION_NAME


# PEP 249 module globals
apilevel = '2.0'
threadsafety = 2  # Threads may share the module and connections.
paramstyle = 'named'  # Python extended format codes, e.g. ...WHERE name=%(name)s


class Error(Exception):
    pass


class State(enum.Enum):
    NONE = 0
    RUNNING = 1
    FINISHED = 2


def connect(*args, **kwargs):
    """Constructor for creating a connection to the database. See class :py:class:`Connection` for
    arguments.
    :returns: a :py:class:`Connection` object.
    """
    return Connection(*args, **kwargs)


FALLBACK_POLICIES = {
    "unsupported": ["ODPS-185"],
    "upgrading": ["ODPS-182", "ODPS-184"],
    "noresource": ["ODPS-183"],
    "timeout": ["ODPS-186"],
    "generic": ["ODPS-180"]
}

FALLBACK_POLICY_ALIASES = {
    "default": ["unsupported", "upgrading", "noresource", "timeout"],
    "all": ["unsupported", "upgrading", "noresource", "timeout", "generic"]
}


class Connection(object):
    def __init__(self, access_id=None, secret_access_key=None, project=None,
                 endpoint=None, session_name=None, odps=None, hints=None, **kw):
        if isinstance(access_id, ODPS):
            access_id, odps = None, access_id

        if odps is None:
            # pop unsupported
            kw.pop("use_sqa", None)
            kw.pop("fallback_policy", None)
            self._odps = ODPS(access_id=access_id, secret_access_key=secret_access_key,
                              project=project, endpoint=endpoint, **kw)
        else:
            if access_id is not None:
                raise ValueError('Either access_id or odps can be specified')
            self._odps = odps
        self._session_name = PUBLIC_SESSION_NAME
        if session_name is not None:
            self._session_name = session_name
        self._use_sqa = (kw.pop('use_sqa', False) != False)
        self._fallback_policy = kw.pop('fallback_policy', '')
        self._project_as_schema = kw.pop(
            'project_as_schema', options.sqlalchemy.project_as_schema
        )
        self._hints = hints

    @property
    def odps(self):
        return self._odps

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def cursor(self, *args, **kwargs):
        """Return a new :py:class:`Cursor` object using the connection."""
        return Cursor(
            self, *args, use_sqa=self._use_sqa,
            fallback_policy=self._fallback_policy,
            hints=self._hints, **kwargs
        )

    def close(self):
        # there is no long polling for ODPS
        # do nothing for close
        pass

    def commit(self):
        # ODPS does not support transactions
        # do nothing for commit
        pass

    def rollback(self):
        raise NotSupportedError("ODPS does not have transactions")


default_arraysize = 1000


class Cursor(object):
    def __init__(self, connection, arraysize=default_arraysize,
            use_sqa=False, fallback_policy='', hints=None, **kwargs):
        self._connection = connection
        self._arraysize = arraysize
        self._reset_state()
        self.lastrowid = None
        self._use_sqa = use_sqa
        self._fallback_policy = []
        self._hints = hints
        fallback_policies = map(lambda x: x.strip(), fallback_policy.split(','))
        for policy in fallback_policies:
            if policy in FALLBACK_POLICY_ALIASES:
                self._fallback_policy.extend(FALLBACK_POLICY_ALIASES[policy])
            else:
                self._fallback_policy.append(policy)

    def _reset_state(self):
        self._state = State.NONE
        self._description = None
        # odps instance and download session
        self._instance = None
        self._download_session = None

    @property
    def arraysize(self):
        return self._arraysize

    @property
    def connection(self):
        return self._connection

    @arraysize.setter
    def arraysize(self, value):
        try:
            self._arraysize = max(int(value), default_arraysize)
        except TypeError:
            warnings.warn('arraysize has to be a integer, got {}, '
                          'will set default value 1000'.format(value))
            self._arraysize = default_arraysize

    @property
    def description(self):
        """This read-only attribute is a sequence of 7-item sequences.

        Each of these sequences contains information describing one result column:

        - name
        - type_code
        - display_size (None in current implementation)
        - internal_size (None in current implementation)
        - precision (None in current implementation)
        - scale (None in current implementation)
        - null_ok (always True in current implementation)

        This attribute will be ``None`` for operations that do not return rows or if the cursor has
        not had an operation invoked via the :py:meth:`execute` method yet.

        The ``type_code`` can be interpreted by comparing it to the Type Objects specified in the
        section below.
        """
        if self._instance is None:
            return
        if self._description is None:
            self._check_download_session()
            self._description = []
            if self._download_session is not None:
                for col in self._download_session.schema.columns:
                    self._description.append((
                        col.name, col.type.name,
                        None, None, None, None, True
                    ))
            else:
                self._description.append((
                    '_c0', 'string', None, None,
                    None, None, True
                ))
        return self._description

    @staticmethod
    def escape_string(item):
        item = to_str(item)
        return "'{}'".format(
            item
                .replace('\\', '\\\\')
                .replace("'", "\\'")
                .replace('\r', '\\r')
                .replace('\n', '\\n')
                .replace('\t', '\\t')
        )

    def execute(self, operation, parameters=None, **kwargs):
        """Prepare and execute a database operation (query or command).

        Parameters may be provided as sequence or mapping and will be bound to variables in the
        operation. Variables are specified in a database-specific notation (see the module's
        ``paramstyle`` attribute for details).

        Return values are not defined.
        """
        for k in ['async', 'async_']:
            if k in kwargs:
                async_ = kwargs[k]
                break
        else:
            async_ = False

        # prepare statement
        if parameters is None:
            sql = operation
        else:
            sql = operation
            for origin, replacement in parameters.items():
                if isinstance(replacement, six.string_types):
                    replacement = self.escape_string(replacement)

                pattern_str = ":%s([,)])?" % re.escape(to_str(origin))
                replacement_str = "%s\\1" % to_str(replacement)
                sql = re.sub(pattern_str, replacement_str, to_str(sql))

        self._reset_state()

        odps = self._connection.odps
        run_sql = odps.execute_sql
        if self._use_sqa:
            run_sql = self._run_sqa_with_fallback
        if async_:
            run_sql = odps.run_sql
        hints = self._hints or {}
        hints.update(kwargs.get("hints") or {})
        self._instance = run_sql(sql, hints=hints)

    def executemany(self, operation, seq_of_parameters):
        for parameter in seq_of_parameters:
            self.execute(operation, parameter)

    def _sqa_error_should_fallback(self, err_str):
        if 'ODPS-18' not in err_str:
            return False
        for fallback_case in self._fallback_policy:
            fallback_error = FALLBACK_POLICIES.get(fallback_case, None)
            if fallback_error is None:
                continue
            for error_code in fallback_error:
                if error_code in err_str:
                    return True
        return False

    def _run_sqa_with_fallback(self, sql, **kw):
        odps = self._connection.odps
        session_name = self._connection._session_name
        inst = None
        while True:
            try:
                if inst is None:
                    inst = odps.run_sql_interactive(sql, service_name = session_name)
                else:
                    inst.wait_for_success(interval=0.5)
                rd = inst.open_reader(tunnel=True, limit=False)
                if not rd:
                    raise ODPSError('failed to create direct download')
                rd.schema  # will check if task is ok
                self._download_session = rd
                return inst
            except ODPSError as e:
                if self._sqa_error_should_fallback(str(e)):
                    return odps.execute_sql(sql)
                elif "OdpsTaskTimeout" in str(e):
                    # tunnel failed to wait data cache result. fallback to normal wait.
                    pass
                else:
                    raise e

    def cancel(self):
        if self._instance is not None:
            self._instance.stop()

    def close(self):
        self._reset_state()

    def _check_download_session(self):
        if not self._download_session and self._instance:
            try:
                self._download_session = self._instance.open_reader(
                    tunnel=True, limit=False)
            except InstanceTypeNotSupported:
                # not select, cannot create session
                self._download_session = None

    def _fetch_non_select(self):
        # not select
        # just return reader.raw
        with self._instance.open_reader() as reader:
            return [(reader.raw,)]

    def _fetch(self, size):
        self._check_download_session()

        if self._download_session is None:
            return self._fetch_non_select()

        results = []
        i = 0
        while size == -1 or i < size:
            try:
                results.append(next(self._download_session).values)
            except StopIteration:
                break
            i += 1
        return results

    def __iter__(self):
        while True:
            res = self.fetchone()
            if res is not None:
                yield res
            else:
                break

    def next(self):
        res = self.fetchone()
        if res is not None:
            yield res
        else:
            raise StopIteration

    def fetchone(self):
        self._check_download_session()
        results = self._fetch(1)
        if len(results) == 1:
            return results[0]

    def fetchmany(self, size=None):
        if size is None:
            size = self._arraysize
        return self._fetch(size)

    def fetchall(self):
        return self._fetch(-1)



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.
# http://code.google.com/p/protobuf/
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""This implementation is modified from google's original Protobuf implementation.
The original author is: robinson@google.com (Will Robinson).
Modified by onesuperclark@gmail.com(onesuper).
"""


class Error(Exception):
    pass


class DecodeError(Error):
    pass


class EncodeError(Error):
    pass




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import multiprocessing
import threading
import time
import struct
import os
import math
import locale
import codecs
import warnings

try:
    from collections.abc import Iterable
except ImportError:  # pragma: no cover
    from collections import Iterable

try:
    import fcntl
    import termios
    import signal
    _CAN_RESIZE_TERMINAL = True
except ImportError:
    _CAN_RESIZE_TERMINAL = False

try:
    from IPython import get_ipython
except ImportError:
    pass
try:
    get_ipython()
except NameError:
    OutStream = None
    IPythonIOStream = None
    widgets = None
else:
    from IPython import version_info
    ipython_major_version = version_info[0]

    try:
        from ipykernel.iostream import OutStream
    except ImportError:
        try:
            from IPython.zmq.iostream import OutStream
        except ImportError:
            if ipython_major_version < 4:
                try:
                    from IPython.kernel.zmq.iostream import OutStream
                except ImportError:
                    OutStream = None
            else:
                OutStream = None

    if OutStream is not None:
        from IPython.utils import io as ipyio
        # On Windows in particular this is necessary, as the io.stdout stream
        # in IPython gets hooked up to some pyreadline magic to handle colors
        try:
            IPythonIOStream = ipyio.IOStream
        except AttributeError:
            IPythonIOStream = None
    else:
        OutStream = None
        IPythonIOStream = None

    # On Windows, in IPython 2 the standard I/O streams will wrap
    # pyreadline.Console objects if pyreadline is available; this should
    # be considered a TTY
    try:
        from pyreadyline.console import Console as PyreadlineConsole
    except ImportError:
        # Just define a dummy class
        class PyreadlineConsole(object): pass

    # import widgets and display
    try:
        if ipython_major_version < 4:
            from IPython.html import widgets
        else:
            from ipywidgets import widgets
    except ImportError:
        widgets = None
    from IPython.display import display

    # ignore widgets deprecated message
    def _ignore_deprecated_warnings():
        warnings.filterwarnings('ignore', category=DeprecationWarning, module=r'.*widget.*')

    if get_ipython and get_ipython():
        get_ipython().events.register('pre_execute', _ignore_deprecated_warnings)

from .compat import six


_DEFAULT_ENCODING = 'utf-8'
_initial_defencoding = None


def detect_console_encoding():
    """
    Try to find the most capable encoding supported by the console.
    slighly modified from the way IPython handles the same issue.
    """
    import locale
    global _initial_defencoding

    encoding = None
    try:
        encoding = sys.stdout.encoding or sys.stdin.encoding
    except AttributeError:
        pass

    # try again for something better
    if not encoding or 'ascii' in encoding.lower():
        try:
            encoding = locale.getpreferredencoding()
        except Exception:
            pass

    # when all else fails. this will usually be "ascii"
    if not encoding or 'ascii' in encoding.lower():
        encoding = sys.getdefaultencoding()

    # GH3360, save the reported defencoding at import time
    # MPL backends may change it. Make available for debugging.
    if not _initial_defencoding:
        _initial_defencoding = sys.getdefaultencoding()

    return encoding


def get_terminal_size():
    """
    Detect terminal size and return tuple = (width, height).

    Only to be used when running in a terminal. Note that the IPython notebook,
    IPython zmq frontends, or IDLE do not run in a terminal,
    """
    import platform
    current_os = platform.system()
    tuple_xy = None
    if current_os == 'Windows':
        tuple_xy = _get_terminal_size_windows()
        if tuple_xy is None:
            tuple_xy = _get_terminal_size_tput()
            # needed for window's python in cygwin's xterm!
    if current_os == 'Linux' or \
        current_os == 'Darwin' or \
            current_os.startswith('CYGWIN'):
        tuple_xy = _get_terminal_size_linux()
    if tuple_xy is None:
        tuple_xy = (80, 25)      # default value
    return tuple_xy


def in_interactive_session():
    """ check if we're running in an interactive shell

    returns True if running under python/ipython interactive shell
    """
    def check_main():
        import __main__ as main
        return not hasattr(main, '__file__')

    try:
        return __IPYTHON__ or check_main()  # noqa: F821
    except:
        return check_main()


def in_ipython_frontend():
    """
    check if we're inside an an IPython zmq frontend
    """
    try:
        ip = get_ipython()
        return 'zmq' in str(type(ip)).lower()
    except:
        pass

    return False


_backend_name = None


def get_notebook_backend():
    global _backend_name
    if _backend_name is not None:
        return _backend_name

    if 'VSCODE_PID' in os.environ:
        _backend_name = 'VSCode'
        return _backend_name

    # DSW always use jupyter lab
    for env_name in os.environ.keys():
        if env_name.lower().startswith('dsw_'):
            _backend_name = 'DSW'
            return _backend_name

    for arg in sys.argv:
        if arg.endswith('.json') and os.path.exists(arg):
            ipy_json_path = os.path.dirname(arg)
            # jupyter lab will generate a jpserver-{pid}.json, while
            # jupyter notebook will generate a nbserver-{pid}.json instead
            lab_cfg_path = os.path.join(ipy_json_path, 'jpserver-%d.json' % os.getppid())
            if os.path.exists(lab_cfg_path):
                _backend_name = 'JupyterLab'
                return _backend_name
    _backend_name = 'JupyterNotebook'
    return _backend_name


def is_widgets_available():
    from .config import options

    if not options.display.notebook_widget or widgets is None:
        return False

    # todo when widget for lab or vscode ready, change this
    if get_notebook_backend() in ('DSW', 'JupyterLab', 'VSCode'):
        return False

    if hasattr(widgets.Widget, '_version_validated'):
        return bool(getattr(widgets.Widget, '_version_validated', None))
    else:
        return True


def in_qtconsole():
    """
    check if we're inside an IPython qtconsole

    DEPRECATED: This is no longer needed, or working, in IPython 3 and above.
    """
    try:
        ip = get_ipython()
        front_end = (
            ip.config.get('KernelApp', {}).get('parent_appname', "") or
            ip.config.get('IPKernelApp', {}).get('parent_appname', "")
        )
        if 'qtconsole' in front_end.lower():
            return True
    except:
        return False
    return False


def get_console_size():
    """Return console size as tuple = (width, height).

    Returns (None,None) in non-interactive session.
    """
    from .config import options

    display_width = options.display.width
    # deprecated.
    display_height = options.display.max_rows

    # Consider
    # interactive shell terminal, can detect term size
    # interactive non-shell terminal (ipnb/ipqtconsole), cannot detect term
    # size non-interactive script, should disregard term size

    # in addition
    # width,height have default values, but setting to 'None' signals
    # should use Auto-Detection, But only in interactive shell-terminal.
    # Simple. yeah.

    if in_interactive_session():
        if in_ipython_frontend():
            # sane defaults for interactive non-shell terminal
            # match default for width,height in config_init
            try:
                from pandas.core.config import get_default_val
                terminal_width = get_default_val('display.width')
                terminal_height = get_default_val('display.max_rows')
            except ImportError:
                terminal_width, terminal_height = None, None
        else:
            # pure terminal
            terminal_width, terminal_height = get_terminal_size()
    else:
        terminal_width, terminal_height = None, None

    # Note if the User sets width/Height to None (auto-detection)
    # and we're in a script (non-inter), this will return (None,None)
    # caller needs to deal.
    return (display_width or terminal_width, display_height or terminal_height)


def _get_terminal_size_windows():
    res = None
    try:
        from ctypes import windll, create_string_buffer

        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12

        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
    except:
        return None
    if res:
        import struct
        (bufx, bufy, curx, cury, wattr, left, top, right, bottom, maxx,
         maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
        sizex = right - left + 1
        sizey = bottom - top + 1
        return sizex, sizey
    else:
        return None


def _get_terminal_size_tput():
    # get terminal width
    # src: http://stackoverflow.com/questions/263890/how-do-i-find-the-width
    # -height-of-a-terminal-window
    try:
        import subprocess
        proc = subprocess.Popen(["tput", "cols"],
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE)
        output = proc.communicate(input=None)
        cols = int(output[0])
        proc = subprocess.Popen(["tput", "lines"],
                                stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE)
        output = proc.communicate(input=None)
        rows = int(output[0])
        return (cols, rows)
    except:
        return None


def _get_terminal_size_linux():
    def ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios
            import struct

            cr = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
        except:
            return None
        return cr
    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_GWINSZ(fd)
            os.close(fd)
        except:
            pass
    if not cr or cr == (0, 0):
        try:
            from os import environ as env
            cr = (env['LINES'], env['COLUMNS'])
        except:
            return None
    return int(cr[1]), int(cr[0])


def _get_stdout(stderr=False):
    """
    This utility function contains the logic to determine what streams to use
    by default for standard out/err.

    Typically this will just return `sys.stdout`, but it contains additional
    logic for use in IPython on Windows to determine the correct stream to use
    (usually ``IPython.util.io.stdout`` but only if sys.stdout is a TTY).
    """

    if stderr:
        stream = 'stderr'
    else:
        stream = 'stdout'

    sys_stream = getattr(sys, stream)

    if IPythonIOStream is None:
        return sys_stream

    ipyio_stream = getattr(ipyio, stream)

    if isatty(sys_stream) and isatty(ipyio_stream):
        # Use the IPython console output stream
        return ipyio_stream
    else:
        # sys.stdout was set to some other non-TTY stream (a file perhaps)
        # so just use it directly
        return sys_stream


def isatty(file):
    """
    Returns `True` if `file` is a tty.

    Most built-in Python file-like objects have an `isatty` member,
    but some user-defined types may not, so this assumes those are not
    ttys.
    """
    proc_name = multiprocessing.current_process().name
    try:
        thread_name = threading.current_thread().name
    except AttributeError:  # pragma: no cover
        thread_name = threading.current_thread().getName()

    if proc_name != 'MainProcess' or thread_name != 'MainThread':
        return False

    if hasattr(file, 'isatty'):
        return file.isatty()
    elif (OutStream is not None and
          isinstance(file, (OutStream, IPythonIOStream)) and
          ((hasattr(file, 'name') and file.name == 'stdout') or
           (hasattr(file, 'stream') and
               isinstance(file.stream, PyreadlineConsole)))):
        # File is an IPython OutStream or IOStream and
        #    File name is 'stdout' or
        #    File wraps a Console
        return True
    return False


def _terminal_size(file=None):
    """
    Returns a tuple (height, width) containing the height and width of
    the terminal.

    This function will look for the width in height in multiple areas
    before falling back on the width and height in astropy's
    configuration.
    """
    from .config import options

    if file is None:
        file = _get_stdout()

    try:
        s = struct.pack(str("HHHH"), 0, 0, 0, 0)
        x = fcntl.ioctl(file, termios.TIOCGWINSZ, s)
        (lines, width, xpixels, ypixels) = struct.unpack(str("HHHH"), x)
        if lines > 12:
            lines -= 6
        if width > 10:
            width -= 1
        return (lines, width)
    except:
        try:
            # see if POSIX standard variables will work
            return (int(os.environ.get('LINES')),
                    int(os.environ.get('COLUMNS')))
        except TypeError:
            # fall back on configuration variables, or if not
            # set, (25, 80)
            lines = options.console.max_lines
            width = options.console.max_width
            if lines is None:
                lines = 25
            if width is None:
                width = 80
            return lines, width


def _color_text(text, color):
    """
    Returns a string wrapped in ANSI color codes for coloring the
    text in a terminal::

        colored_text = color_text('Here is a message', 'blue')

    This won't actually effect the text until it is printed to the
    terminal.

    Parameters
    ----------
    text : str
        The string to return, bounded by the color codes.
    color : str
        An ANSI terminal color name. Must be one of:
        black, red, green, brown, blue, magenta, cyan, lightgrey,
        default, darkgrey, lightred, lightgreen, yellow, lightblue,
        lightmagenta, lightcyan, white, or '' (the empty string).
    """
    color_mapping = {
        'black': '0;30',
        'red': '0;31',
        'green': '0;32',
        'brown': '0;33',
        'blue': '0;34',
        'magenta': '0;35',
        'cyan': '0;36',
        'lightgrey': '0;37',
        'default': '0;39',
        'darkgrey': '1;30',
        'lightred': '1;31',
        'lightgreen': '1;32',
        'yellow': '1;33',
        'lightblue': '1;34',
        'lightmagenta': '1;35',
        'lightcyan': '1;36',
        'white': '1;37'}

    if sys.platform == 'win32' and OutStream is None:
        # On Windows do not colorize text unless in IPython
        return text

    color_code = color_mapping.get(color, '0;39')
    return '\033[{0}m{1}\033[0m'.format(color_code, text)


def _decode_preferred_encoding(s):
    """Decode the supplied byte string using the preferred encoding
    for the locale (`locale.getpreferredencoding`) or, if the default encoding
    is invalid, fall back first on utf-8, then on latin-1 if the message cannot
    be decoded with utf-8.
    """

    enc = locale.getpreferredencoding()
    try:
        try:
            return s.decode(enc)
        except LookupError:
            enc = _DEFAULT_ENCODING
        return s.decode(enc)
    except UnicodeDecodeError:
        return s.decode('latin-1')


def _write_with_fallback(s, write, fileobj):
    """Write the supplied string with the given write function like
    ``write(s)``, but use a writer for the locale's preferred encoding in case
    of a UnicodeEncodeError.  Failing that attempt to write with 'utf-8' or
    'latin-1'.
    """

    if IPythonIOStream is not None and isinstance(fileobj, IPythonIOStream):
        # If the output stream is an IPython.utils.io.IOStream object that's
        # not going to be very helpful to us since it doesn't raise any
        # exceptions when an error occurs writing to its underlying stream.
        # There's no advantage to us using IOStream.write directly though;
        # instead just write directly to its underlying stream:
        write = fileobj.stream.write

    try:
        write(s)
        return write
    except UnicodeEncodeError:
        # Let's try the next approach...
        pass

    enc = locale.getpreferredencoding()
    try:
        Writer = codecs.getwriter(enc)
    except LookupError:
        Writer = codecs.getwriter(_DEFAULT_ENCODING)

    f = Writer(fileobj)
    write = f.write

    try:
        write(s)
        return write
    except UnicodeEncodeError:
        Writer = codecs.getwriter('latin-1')
        f = Writer(fileobj)
        write = f.write

    # If this doesn't work let the exception bubble up; I'm out of ideas
    write(s)
    return write


def color_print(*args, **kwargs):
    """
    Prints colors and styles to the terminal uses ANSI escape
    sequences.

    ::

       color_print('This is the color ', 'default', 'GREEN', 'green')

    Parameters
    ----------
    positional args : str
        The positional arguments come in pairs (*msg*, *color*), where
        *msg* is the string to display and *color* is the color to
        display it in.

        *color* is an ANSI terminal color name.  Must be one of:
        black, red, green, brown, blue, magenta, cyan, lightgrey,
        default, darkgrey, lightred, lightgreen, yellow, lightblue,
        lightmagenta, lightcyan, white, or '' (the empty string).

    file : writeable file-like object, optional
        Where to write to.  Defaults to `sys.stdout`.  If file is not
        a tty (as determined by calling its `isatty` member, if one
        exists), no coloring will be included.

    end : str, optional
        The ending of the message.  Defaults to ``\\n``.  The end will
        be printed after resetting any color or font state.
    """
    from .config import options

    file = kwargs.get('file', _get_stdout())

    end = kwargs.get('end', '\n')

    write = file.write
    if isatty(file) and options.console.use_color:
        for i in range(0, len(args), 2):
            msg = args[i]
            if i + 1 == len(args):
                color = ''
            else:
                color = args[i + 1]

            if color:
                msg = _color_text(msg, color)

            # Some file objects support writing unicode sensibly on some Python
            # versions; if this fails try creating a writer using the locale's
            # preferred encoding. If that fails too give up.
            if not six.PY3 and isinstance(msg, bytes):
                msg = _decode_preferred_encoding(msg)

            write = _write_with_fallback(msg, write, file)

        write(end)
    else:
        for i in range(0, len(args), 2):
            msg = args[i]
            if not six.PY3 and isinstance(msg, bytes):
                # Support decoding bytes to unicode on Python 2; use the
                # preferred encoding for the locale (which is *sometimes*
                # sensible)
                msg = _decode_preferred_encoding(msg)
            write(msg)
        write(end)


def human_time(seconds):
    """
    Returns a human-friendly time string that is always exactly 6
    characters long.

    Depending on the number of seconds given, can be one of::

        1w 3d
        2d 4h
        1h 5m
        1m 4s
          15s

    Will be in color if console coloring is turned on.

    Parameters
    ----------
    seconds : int
        The number of seconds to represent

    Returns
    -------
    time : str
        A human-friendly representation of the given number of seconds
        that is always exactly 6 characters.
    """
    units = [
        ('y', 60 * 60 * 24 * 7 * 52),
        ('w', 60 * 60 * 24 * 7),
        ('d', 60 * 60 * 24),
        ('h', 60 * 60),
        ('m', 60),
        ('s', 1),
    ]

    seconds = int(seconds)

    if seconds < 60:
        return '   {0:2d}s'.format(seconds)
    for i in range(len(units) - 1):
        unit1, limit1 = units[i]
        unit2, limit2 = units[i + 1]
        if seconds >= limit1:
            return '{0:2d}{1}{2:2d}{3}'.format(
                seconds // limit1, unit1,
                (seconds % limit1) // limit2, unit2)
    return '  ~inf'


def human_file_size(size):
    """
    Returns a human-friendly string representing a file size
    that is 2-4 characters long.

    For example, depending on the number of bytes given, can be one
    of::

        256b
        64k
        1.1G

    Parameters
    ----------
    size : int
        The size of the file (in bytes)

    Returns
    -------
    size : str
        A human-friendly representation of the size of the file
    """
    suffixes = ' kMGTPEZY'
    if size == 0:
        num_scale = 0
    else:
        num_scale = int(math.floor(math.log(size) / math.log(1000)))
    num_scale = max(num_scale, 0)
    if num_scale >= len(suffixes):
        suffix = '?'
    else:
        suffix = suffixes[num_scale]
    num_scale = int(math.pow(1000, num_scale))
    value = float(size) / num_scale
    str_value = str(value)
    if suffix == ' ':
        if '.' in str_value:
            str_value = str_value[:str_value.index('.')]
    elif str_value[2] == '.':
        str_value = str_value[:2]
    else:
        str_value = str_value[:3]
    return "{0:>3s}{1}".format(str_value, suffix)


def create_progress_widget():
    # Import only if an IPython widget, i.e., widget in iPython NB
    if ipython_major_version < 4:
        widget_cls = widgets.FloatProgressWidget
        from IPython.utils.traitlets import Unicode
    else:
        widget_cls = widgets.FloatProgress
        from traitlets import Unicode

    from .ui.common import build_trait

    class TransientProgressBar(widget_cls):
        _view_name = build_trait(Unicode, 'TransientProgressView', sync=True)
        _view_module = build_trait(Unicode, 'pyodps/progress', sync=True)

    return TransientProgressBar()


class ProgressBar(six.Iterator):
    """
    A class to display a progress bar in the terminal.

    It is designed to be used either with the ``with`` statement::

        with ProgressBar(len(items)) as bar:
            for item in enumerate(items):
                bar.update()

    or as a generator::

        for item in ProgressBar(items):
            item.process()
    """
    def __init__(self, total_or_items, ipython_widget=False, file=None):
        """
        Parameters
        ----------
        total_or_items : int or sequence
            If an int, the number of increments in the process being
            tracked.  If a sequence, the items to iterate over.

        ipython_widget : bool, optional
            If `True`, the progress bar will display as an IPython
            notebook widget.

        file : writable file-like object, optional
            The file to write the progress bar to.  Defaults to
            `sys.stdout`.  If `file` is not a tty (as determined by
            calling its `isatty` member, if any, or special case hacks
            to detect the IPython console), the progress bar will be
            completely silent.
        """
        if ipython_widget:
            # Import only if ipython_widget, i.e., widget in IPython
            # notebook
            try:
                if ipython_major_version < 4:
                    from IPython.html import widgets  # noqa: F401
                else:
                    from ipywidgets import widgets  # noqa: F401
                from IPython.display import display  # noqa: F401

                ipython_widget = is_widgets_available()
            except ImportError:
                ipython_widget = False

        if file is None:
            file = _get_stdout()

        if not isatty(file) and not ipython_widget:
            self.update = self._silent_update
            self._silent = True
        else:
            self._silent = False

        if isinstance(total_or_items, Iterable):
            self._items = iter(total_or_items)
            self._total = len(total_or_items)
        else:
            try:
                self._total = int(total_or_items)
            except TypeError:
                raise TypeError("First argument must be int or sequence")
            else:
                self._items = iter(range(self._total))

        self._file = file
        self._start_time = time.time()
        self._human_total = human_file_size(self._total)
        self._ipython_widget = ipython_widget


        self._signal_set = False
        if not ipython_widget:
            self._should_handle_resize = (
                _CAN_RESIZE_TERMINAL and self._file.isatty())
            self._handle_resize()
            if self._should_handle_resize:
                signal.signal(signal.SIGWINCH, self._handle_resize)
                self._signal_set = True

        self.update(0)

    def _handle_resize(self, signum=None, frame=None):
        terminal_width = _terminal_size(self._file)[1]
        self._bar_length = terminal_width - 37

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self._silent:
            if exc_type is None:
                self.update(self._total)
            self._file.write('\n')
            self._file.flush()
            if self._signal_set:
                signal.signal(signal.SIGWINCH, signal.SIG_DFL)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            rv = next(self._items)
        except StopIteration:
            self.__exit__(None, None, None)
            raise
        else:
            self.update()
            return rv

    def update(self, value=None):
        """
        Update progress bar via the console or notebook accordingly.
        """

        # Update self.value
        if value is None:
            value = self._current_value + 1
        self._current_value = value

        # Choose the appropriate environment
        if self._ipython_widget:
            try:
                self._update_ipython_widget(value)
            except RuntimeError:
                pass
        else:
            self._update_console(value)

    def close(self):
        if not self._ipython_widget:
            self.__exit__(None, None, None)
        if self._ipython_widget and self._widget:
            self._widget.close()

    def _update_console(self, value=None):
        """
        Update the progress bar to the given value (out of the total
        given to the constructor).
        """

        if self._total == 0:
            frac = 1.0
        else:
            frac = float(value) / float(self._total)

        file = self._file
        write = file.write

        if frac > 1:
            bar_fill = int(self._bar_length)
        else:
            bar_fill = int(float(self._bar_length) * frac)
        write('\r|')
        color_print('=' * bar_fill, 'blue', file=file, end='')
        if bar_fill < self._bar_length:
            color_print('>', 'green', file=file, end='')
            write('-' * (self._bar_length - bar_fill - 1))
        write('|')

        if value >= self._total:
            t = time.time() - self._start_time
            prefix = '     '
        elif value <= 0:
            t = None
            prefix = ''
        else:
            t = ((time.time() - self._start_time) * (1.0 - frac)) / frac
            prefix = ' ETA '
        write(' {0:>4s}/{1:>4s}'.format(
            human_file_size(value),
            self._human_total))
        write(' ({0:>6s}%)'.format('{0:.2f}'.format(frac * 100.0)))
        write(prefix)
        if t is not None:
            write(human_time(t))
        self._file.flush()

    def _update_ipython_widget(self, value=None):
        """
        Update the progress bar to the given value (out of a total
        given to the constructor).

        This method is for use in the IPython notebook 2+.
        """

        # Create and display an empty progress bar widget,
        # if none exists.
        if not hasattr(self, '_widget'):
            self._widget = create_progress_widget()
            if in_ipython_frontend() and is_widgets_available():
                display(self._widget)
            self._widget.value = 0

        # Calculate percent completion, and update progress bar
        percent = (float(value)/self._total) * 100.0
        self._widget.value = percent
        self._widget.description =' ({0:>6s}%)'.format('{0:.2f}'.format(percent))

    def _silent_update(self, value=None):
        pass

    @classmethod
    def map(cls, function, items, multiprocess=False, file=None):
        """
        Does a `map` operation while displaying a progress bar with
        percentage complete.

        ::

            def work(i):
                print(i)

            ProgressBar.map(work, range(50))

        Parameters
        ----------
        function : function
            Function to call for each step

        items : sequence
            Sequence where each element is a tuple of arguments to pass to
            *function*.

        multiprocess : bool, optional
            If `True`, use the `multiprocessing` module to distribute each
            task to a different processor core.

        file : writeable file-like object, optional
            The file to write the progress bar to.  Defaults to
            `sys.stdout`.  If `file` is not a tty (as determined by
            calling its `isatty` member, if any), the scrollbar will
            be completely silent.
        """

        results = []

        if file is None:
            file = _get_stdout()

        with cls(len(items), file=file) as bar:
            step_size = max(200, bar._bar_length)
            steps = max(int(float(len(items)) / step_size), 1)
            if not multiprocess:
                for i, item in enumerate(items):
                    results.append(function(item))
                    if (i % steps) == 0:
                        bar.update(i)
            else:
                p = multiprocessing.Pool()
                for i, result in enumerate(
                    p.imap_unordered(function, items, steps)):
                    bar.update(i)
                    results.append(result)
                p.close()
                p.join()

        return results


class StatusLine(object):
    def __init__(self, ipython_widget=False):
        if widgets is None:
            ipython_widget = False

        self.ipython_widget = ipython_widget
        self._widget = None

    def update(self, text):
        if self.ipython_widget:
            if not self._widget:
                self._widget = widgets.HTML()
                if is_widgets_available():
                    display(self._widget)
            self._widget.value = text

    def close(self):
        if self._widget:
            self._widget.close()
